/* Common definitions shared by all C code in this driver */

#ifndef _HPT_DEFS_H
#define _HPT_DEFS_H

/* tuning. currently, SOFTINTR is used for "production".
 *  no intr stuff will work nicely for debug.
 * HARDINTR will unfortunately blow the system up. seems to be
 * a bad interaction between solaris and my BP6
 */

/* define 0 or 1 of these, NOT BOTH */
/*#define USE_SOFTINTR 1*/

#define USE_HARDINTR 1 /* THis is for actual physical interrupts! */

#if !defined(USE_SOFTINTR) && !defined(USE_HARDINTR)
# define NO_INTR 1
#endif

/* Do NOT include these in any other files!!! */
/* Also, because of these includes, make sure that defs.h is */
/* included AFTER ANY OTHER system includes!! */

/* since varargs is BROKEN WITH GCC, arrgg... */
/*
#define DLPRINT(minlev, msg) if(debug_level>=minlev){cmn_err(CE_NOTE, msg);}
#define DLPRINT2(minlev, f,a) if(debug_level>=minlev){cmn_err(CE_NOTE, f,a);}
#define DLPRINT3(minlev, f,a,b) if(debug_level>=minlev){cmn_err(CE_NOTE, f,a,b);}
*/
#define DLPRINT(minlev, msg) if(debug_level>=minlev){printf(msg);}
#define DLPRINT2(minlev, f,a) if(debug_level>=minlev){printf(f,a);}
#define DLPRINT3(minlev, f,a,b) if(debug_level>=minlev){printf(f,a,b);}



#define HPT_FULLDISK 0x1000 /* flag to denote fdisk partitions, not slices. */

/* This is "drives per instance */
#define HPT_NUMBUSES 1  /* highpoint has 1, intel has 2 */
#define HPT_NUMDRIVES (2 * HPT_NUMBUSES)

#define MAXTAG 32  /* maximum queue depth for "tagged" commands */

/* The idea for minor numbers is:
 * 0x00-0x0f = slices 0-15, first disk
 * 0x10-0x1f = slices 0-15, second disk
 * 0x20-0x2f = slices 0-15, second disk
 * 0x30-0x3f = slices 0-15, second disk
 * 0x40 : start of next instance usage
 * 0x100:  "whole disk", including MBR/fdisk partition info
 * 0x101:  FDISK partition #1
 */
#define DEV_SLICEMASK  0x0f /* bits for slice in a drive, in minor num */
#define DEV_DRIVEMASK  0x30 /* bits for drive. 4 per instance */
#define DEV_DRIVEMULT  0x10 /* bits for drive. 4 per instance */
#define DEV_INSTANCEMULT 0x40
#define GETINSTANCE(minor)  ((minor&0xfff) >> 6)

/* XXX THis is stolen from isp.h.
 * It should really be updated, methinks.
 * I added BURST32
 */
/* 0x3f means all burst-sizes up to 32 bits supported */
/* 0x0f would mean all sizes up to 8 bits supported */

#define        HPT_BURSTSIZES		0x0f
/*#define        HPT_BURSTSIZES		0x10*/
/*#define        HPT_BURSTSIZES		0x3f*/
	/* 16-bits of addrspace is max sgtable entry size*/

#define HPT_SECTSIZE 512 /* sector size */

/* yes, it's silly. but the limit of a single UDMA transfer is
 * actually 64k, or something. Whereas IDE can theoreticaly grab
 * 128k at a time.
 */
#define MAXIDETRANSFER 0x20000	/* in bytes, max transfer. (128k) */
#define MAXDMATRANSFER 0xe000 /* in bytes, max transfer. (56k)(0xe000) */

/* well, everyone else sets this to 1, so... */
#define HPT_MIN_XFER 1

/* These are status flags for the IDE status/command register */
#define	IDESTATUS_ERR		0x01
/* 0x02 obsolete*/
/* 0x04 obsolete*/
#define	IDESTATUS_DATAREADY	0x08 /* aka "DRQ" bit */
#define	IDESTATUS_SERVICE	0x10 /* waiting for SERVICE cmd */
#define	IDESTATUS_DF		0x20 /* device FAULT */
#define	IDESTATUS_DEVREADY	0x40 /* aka "DRDY" bit */
#define	IDESTATUS_BUSY		0x80


/* Here are a list of commands from the ATA5 spec. */
/* But just the ones we care about */
#define IDECMD_NOP		0x00 /* this DOES generate an interrupt!*/
#define IDECMD_DEVRESET		0x08 /* device reset */
#define IDECMD_READ		0x20
#define IDECMD_READDMA		0xc8
#define IDECMD_READDMAQUEUED	0xc7
#define IDECMD_WRITE		0x30
#define IDECMD_WRITEDMA		0xca
#define IDECMD_WRITEDMAQUEUED	0xcc
#define IDECMD_SEEK		0x70
#define IDECMD_FLUSHCACHE	0xe7
#define IDECMD_IDENTIFY		0xec
#define IDECMD_SERVICE		0xa2
#define IDECMD_SETFEATURES	0xef	 /* S8.37 */
#define IDECMD_SLEEP		0xe6

#define IDECONTROL_NOIRQ	0x02
#define IDECONTROL_SRESET	0x04

/* These are used with IDECMD_SETFEATURES */
/* They go in "feature" register */
#define IDE_FEAT_TRANSMODE	0x03
#define IDE_FEAT_READAHEAD	0xaa
#define IDE_FEAT_NOREADAHEAD	0x55
#define IDE_FEAT_SERVICEINT	0x5e/* no SERVICE-flagged interrupt */
#define IDE_FEAT_NOSERVICEINT	0xde
#define IDE_FEAT_WRITECACHE	0x02
#define IDE_FEAT_NOWRITECACHE	0x82
#define IDE_FEAT_RELINT		0x5d /* no SERVICE-flagged interrupt */
#define IDE_FEAT_NORELINT	0xdd /* no RELEASE interrupt */

/* The "drive_head" thing from "commandreg" above is rather ugly.
 * its bits are as follows
 * 7	"always 1"
 * 6	LBA mode enable
 * 5	"always 1"
 * 4	drive select
 * 3-0	"always zeros".
 *
 * The above is from the linux drivers. So I'm not sure about the 3-0 spec.
 * the official dma spec says the full sector value shoudl be 28 bits, not 24.
 * And oh look, those 4 bits would be the upper 24+4.
 */

#define LBA_BASE 0x40	/* The minimum to put the IDE command reg into LBA
			 * mode. This goes in the "drive/head" register
			 */
#define HEAD_CHANNEL2 0x10  /* select secondary drive */

/*
 * This is for UDMA purposes
 * It really belongs in hpt-udma.h, but for now we keep it in here,
 * because stuff in softstate references it.
 *
 * This is a cluster of all the critical info we need to
 * group together, to alloc/dealloc a scatter/gather table
 */
struct sgclump {
					/* If multiple bufs chained, only
					 * the first one will have valid table.
					 * Everyone else will only have
					 * valid bufhandle
					 */
	struct buf *bufptr;		/* kernel struct for user request */
	ddi_dma_handle_t bufhandle;	/* DMA handle for buf */

	struct sg_entry *tableptr;	/* kernel address of table */
	ddi_dma_handle_t sg_handle;	/* allocation handle for Scatter/G. */
	ddi_acc_handle_t sg_dma_access;	/* "access" handle for Scatter/Gather*/
	ddi_dma_cookie_t sg_cookie;	/* cookie for phys addr of table */

	uint32_t physaddr;		/* physaddr start of table */
					/* (when we hand-align)  */
	struct sgclump *next;		/* For when we merge bufs into
					 * 1 I/O request
					 */

	int retry_count;		/* yeah, wasting 3 bytes, but
					 * I like this word-aligned
					 */
};


/* Okay, this is ugly, but I dont remember the wierd C syntax.
 * sue me
 */
typedef struct buf * BUFPTR;

/************************************************************************
 *This struct holds everything we care about on a per-instance
 * basis of the driver.
 * This is what is stored by "ddi_get_soft_state()"
 *  1 softinfo struct == 1 bus. Note that there are usually 2 busses
 *    per chip,
 * I think.
 * So there will probably always be two instances?
 **************************************************************************/

struct hpt_softstate {

	dev_info_t *dip; /* this exists for getinfo(), and finding
			  * our instance number if neccessary
			  */

	ddi_softintr_t softintr_id;

	/* Mutexes are currently handled as follows:
	 * You should pick a channel, and grab the mutex for that,
	 * channel, AFTER getting the instance mutex. EG:
	 * mutex_enter(&instance_mutex);
	 * mutex_enter(&chan1_mutex);
	 * mutex_exit(&instance_mutex);
	 *  [ do stuff here ]
	 * mutex_exit(&chan1_mutex);
	 */
	 
	kmutex_t instance_mutex;
	int8_t detach_start;

	kmutex_t waitqueue_mutex[HPT_NUMDRIVES]; /* adjustments to waitqueue[]*/
	
	kmutex_t activequeue_mutex[HPT_NUMBUSES]; /* for currently-active queue adjust */
						  /* except we use sema now*/

	kmutex_t bus_mutex[HPT_NUMBUSES]; /* for actual IDE/UDMA commands */
	
	kcondvar_t irq_wait[HPT_NUMBUSES]; /* not used right now*/
	kmutex_t wait_mutex[HPT_NUMBUSES]; /* not used right now*/

	ksema_t  activequeue_sema[HPT_NUMBUSES];
	ksema_t  timeout_sema[HPT_NUMBUSES]; /* One timeout thread at a time*/
					     /* used if using softintr */

	timeout_id_t timeout_id[HPT_NUMBUSES]; /* if using timeout() */

	uint8_t queued_ok[HPT_NUMBUSES]; /* Bus can use tagged queueing */

/* This section is all "per bus", really 
 * so for HPT, it isnt realy appropriate. but Im trying to stay
 * compatible with my bmi drivers
 */


/* end of bus-specific section */


	ddi_acc_handle_t pci_conf_h;
	ddi_acc_handle_t hpt_udmaregs_handle;
	struct hpt_udmaregs *hpt_udmaregs; /* udma IO regs */

	BUFPTR tag_array[256];/* for tagged commands*/

	ddi_acc_handle_t		hpt_ide_command_handle[HPT_NUMBUSES];
	struct hpt_ide_commandreg	*ide_command_regs[HPT_NUMBUSES];
	ddi_acc_handle_t		hpt_ide_control_handle[HPT_NUMBUSES];
	struct hpt_ide_controlreg	*ide_control_reg[HPT_NUMBUSES];

	/* interrupt "cookie" for system DMA stuff*/
	ddi_iblock_cookie_t interrupt_cookie;

	/* this is from the GET_FEATURE thing */
	uint16_t	dev_info[HPT_NUMDRIVES][256];
	/* THis is a cached version of what we set the udma speed to */
	int8_t		drive_speed[HPT_NUMDRIVES];/*MODE_UDMAXX*/
	/* when a new request is queued up, retry count is set to non-zero*/
	uint8_t		retry_count[HPT_NUMDRIVES];

	
	uint8_t		blkspersect[HPT_NUMDRIVES];
	
	struct diskhd	waitqueue[HPT_NUMDRIVES]; /* IO requests wait here
						   * until disk is free
						   */

	struct sgclump * activequeue[HPT_NUMBUSES];
	/* When bus is busy, point to what we're doing. Otherwise,
	 * will be NULL
	 */

	/* for tagged commands.. except I'm going to rewrite this*/
	struct sgclump	tag_tables[HPT_NUMDRIVES][MAXTAG];

	/* If FD_NUMPART isnt defined, you need to include
	 *  <sys/dktp/fdisk.h>
	 * before including defs.h
	 *
	 * Remember when accessing, that hpt0D0f1, is
	 * actually fdisklabel[0][==0==]
	 * Dont forget to subtract 1!
	 */
	struct ipart fdisklabel[HPT_NUMDRIVES][FD_NUMPART];

	/* First block, of first solaris fdisk partition. kinda
	 * Just add this to whatever the solaris slice startblock says,
	 * and you have the raw disk sector it actually starts on
	 */
	int sol_startblk[HPT_NUMDRIVES]; 

	/* copy of solaris "partition" infostructs (from sol_startblk) */
	/* You need to have included <sys/vtoc.h> */
	struct vtoc partitions[HPT_NUMDRIVES];

	/* remember, "NDKMAP" is # of slices */


	/* The following is partition information
	 * "fp" means fdisk partition.
	 * The first 16 [0-15] are for disk #1.
	 * The second 16 are for disk #2, aka "channel 2".
	 * So the minor number of a device node is a direct index
	 * into this.
	 *
	 * fp_numblocks is the number of blocks in a particular fdisk
	 * partition
	 * fp_startblock is the ABSOLUTE BLOCK NUMBER on the disk,
	 * of the first block in that partition.
	 */
	uint32_t fp_numblocks[32];
	uint32_t fp_startblock[32];
};


/***********************************************************************
 * Below here  are the actual driver-custom struct definitions,
 * mainly pertaining to register organization
 ***********************************************************************/



/**********************************************************************/
/* Here starts the "Legacy ISA" stuff                                 */

/* see http://developer.intel.com/design/intarch/techinfo/440BX/idefunc.htm,
 * and other places
 */

/* This is for the legacy ISA chipset */
/* This is memory mapped. do not change anything */
struct hpt_ide_commandreg{
	uint8_t data;
	uint8_t error; /* read for error, write for "feature" */
	uint8_t sector_count;
	uint8_t sector_number;
	uint8_t cylinder_low;
	uint8_t cylinder_high;
	uint8_t drive_head; /* see below */
	uint8_t command; /* this is actually status/command */

	/* note: "altstatus"/device control", in next struct */
};

/* NOTE! BMI (440bx) driver has a SINGLE UINT8 mapped by the PCI stuff.
 * But the hpt chipset is rather different
 */
struct hpt_ide_controlreg{
	uint8_t reserved1,reserved2;
	uint8_t control; /* 'alt' status, AND "device control" */
	uint8_t floppy; /* Gah?!! */
};

#define IDE_ERR_ABT	0x04 /* for error reg, above */



#endif /* _HPT_DEFS_H */
