/* I'm trying to keep this file to be JUST UDMA functions. Things
 * that actually are the heart of udma *operation*
 *  (Plus tagged stuff)
 *
 * Remember that the udma "engine" is separate from the regular stuff.
 *
 * But for right now, I'm only moving the "QUEUED" (tagged) operations here.
 * @(#) hpt-udma.c 1.73@(#)
 */

#include <sys/errno.h>
#include <sys/conf.h>
#include <sys/modctl.h>
#include <sys/open.h>
#include <sys/stat.h>
#include <sys/kmem.h>
#include <sys/ksynch.h>

/* THis stuff is for struct softstate, not because we need it here */
#include <sys/vtoc.h>

#include <sys/dktp/fdisk.h> /* just for defs.h */

#include <sys/ddi.h>
#include <sys/sunddi.h>

#include "defs.h"
#include "hpt.h"
#include "hpt-queues.h"
#include "hpt-ide.h"
#include "hpt-udma.h"

/*
 * This is for the scatter/gather table handle.
 * We actually COPY THIS to a tmp struct, whenever
 * we do ddi_dma_alloc_handle().
 */
static ddi_dma_attr_t sg_attributes = {
	DMA_ATTR_V0,
	(unsigned long long)0,			/* dma_attr_addr_lo */
#ifdef RESTRICT_ADDRS
	/*sigh.. im desparate, try anything now */
	/* But this doesnt help the WRITE DMA errors*/
	(unsigned long long)0x0bffffff,		/* dma_attr_addr_hi */
#else
	(unsigned long long)0xffffffff,		/* dma_attr_addr_hi */
#endif
	(unsigned long long)0x0000ffff,		/* dma_attr_count_max */
	(unsigned long long)4096,			/* dma_attr_align */
	HPT_BURSTSIZES ,			/* dma_attr_burstsizes */
	HPT_MIN_XFER,				/* dma_attr_minxfer */
	(unsigned long long)0x00010000,		/* dma_attr_maxxfer */
	(unsigned long long)0x0000ffff,		/* dma_attr_seg */
	HPT_MAX_SG_ENT,				/* dma_attr_sgllen */
	HPT_SECTSIZE,				/* dma_attr_granular */
#ifdef USEPHYS
	DDI_DMA_FORCE_PHYSICAL			/* dma_attr_flags */
#else
	0			/* dma_attr_flags */
#endif
};

/* This is for binding the user buf address to device memory. 
 * These values get copied to a tmp struct for bind purposes.
 */
static ddi_dma_attr_t buf_bind_attributes = {
	DMA_ATTR_V0,
	(unsigned long long)0,			/* dma_attr_addr_lo */
	(unsigned long long)0x7fffffff,		/* dma_attr_addr_hi */
	(unsigned long long)0x0000ffff,		/* dma_attr_count_max */
#ifdef SIMPLE_BUF_ALIGN
	(unsigned long long)4,			/* dma_attr_align */
#else
	/* try aligning to 4k page */
	(unsigned long long)4096,		/* dma_attr_align */
#endif
	HPT_BURSTSIZES ,			/* dma_attr_burstsizes */
	HPT_MIN_XFER,				/* dma_attr_minxfer */
	(unsigned long long)65536,		/* dma_attr_maxxfer */
	(unsigned long long)0x0000ffff,		/* dma_attr_seg */
	HPT_MAX_SG_ENT,				/* dma_attr_sgllen */
	HPT_SECTSIZE,				/* dma_attr_granular */
#ifdef USEPHYS
	DDI_DMA_FORCE_PHYSICAL			/* dma_attr_flags */
#else
	0			/* dma_attr_flags */
#endif
};


/* print to console, the current UDMA status register info*/
/* This should be called with the value of the "Bus Master Status Register" */
void printudmastatus(uint8_t reg8){
	char buf1[50],buf2[50];
	buf1[0]='\0'; buf2[0]='\0';

	if(reg8 & UDMASTATUS_ERR){
		strcpy(buf2,buf1);
		sprintf(buf1,"%s DMA_ERR",buf2);
	}
	if(reg8 & UDMASTATUS_BUSY){
		strcpy(buf2,buf1);
		sprintf(buf1,"%s BUSY",buf2);
	}
	if(reg8 & UDMASTATUS_IRQ){
		strcpy(buf2,buf1);
		sprintf(buf1,"%s IRQ",buf2);
	}
	printf("  print(udma)status: %s", buf1);
}


/**********************************************************************
 * All the Scatter/Gather table routines start here                   *
 **********************************************************************/

static void cleanup_dma_mem(ddi_dma_handle_t *handleptr,
		     ddi_acc_handle_t *acc_handleptr){
	ddi_dma_unbind_handle(*handleptr);
	ddi_dma_mem_free(acc_handleptr);
	ddi_dma_free_handle(handleptr);
}

/* Give me a sg_clump to hold things, and I'll
 * allocate the dma 'HANDLES' for the s/g table, plus the
 * actual table itself.
 * We only allocate memory for the scatter/gather table, we dont
 * populate it with any information.
 *
 * Note that the resulting table must be aligned to a **64K** boundary.
 * Waitaminit.. the intel notes just say that it cannot cross a 4k boundary.
 * So we'll go with that.
 * 
 * The CONTENTS [slots] of the sg_table get filled out at the last minute,
 * in startdma(), with pointers to the user buffer space.
 *
 * return 0 on okay, 1 on fail
 */

int alloc_sg_table(struct hpt_softstate *softstate, struct sgclump *sg_clump){

	dev_info_t *dip = softstate->dip;
	ddi_dma_attr_t tmp_dma_attr;
	size_t padded_len; /* throwaway */
	u_int ccount;
#ifdef UGLYHACK
	uint32_t paddr;
#endif
	tmp_dma_attr = sg_attributes;

	/* Id LIKE to just do this for 64k align.
	 * but ddi_dma_mem_alloc() fails on it
	 */
	/*tmp_dma_attr.dma_attr_align=0x10000;*/
	
	DLPRINT(2,"doing sgalloc magic to 4k align only");
	/* align to 4k page */
	/*tmp_dma_attr.dma_attr_align=0x1000;*/
	

	if(ddi_dma_alloc_handle(dip, &tmp_dma_attr,
	    DDI_DMA_SLEEP, NULL, &(sg_clump->sg_handle))
	    != DDI_SUCCESS) {
		cmn_err(CE_WARN,"failed ddi_dma_alloc_handle");
		return 1;
	}
	/* Allocate the 'table' */
	if (ddi_dma_mem_alloc(sg_clump->sg_handle,
			      (HPT_MAX_SG_ENT+1)*sizeof(SGENTSIZE),
			      &dev_attr, DDI_DMA_CONSISTENT, DDI_DMA_SLEEP,
			      NULL,
			      (caddr_t*)&(sg_clump->tableptr),
			      &padded_len,
			      &(sg_clump->sg_dma_access) )
				!= DDI_SUCCESS) {
		cmn_err(CE_WARN,"alloc_sg_table: failed ddi_dma_mem_alloc");
		return 1;
	}
	/* bind handle to table */
	if (ddi_dma_addr_bind_handle(sg_clump->sg_handle, NULL,
			     (caddr_t)sg_clump->tableptr,
			     SGTABLESIZE,
			     DDI_DMA_WRITE|DDI_DMA_CONSISTENT,
			     DDI_DMA_SLEEP, NULL,
			     &(sg_clump->sg_cookie), &ccount)
					!= DDI_DMA_MAPPED) {

		cmn_err(CE_WARN,
			"alloc_sg_table: failed ddi_dma_addr_bind_handle");
		ddi_dma_free_handle(sg_clump->sg_handle);
		return 1;	
	}

	if(ccount!=1){
		cmn_err(CE_WARN,"hpt: alloc_sg_table() has cookiecount >1");
		cleanup_dma_mem(&(sg_clump->sg_handle),
				&(sg_clump->sg_dma_access));
		return 1;
	}

	sg_clump->physaddr=sg_clump->sg_cookie.dmac_address;
	DLPRINT3(1,"physaddr of sgtable is 0x%x, kaddr=0x%p\n",
		 sg_clump->physaddr,
		 sg_clump->tableptr);
	return 0;

}

/* The counterpart to alloc_sg_table.
 * Will unbind and deallocate the access handle for the SG table.
 *   (access handle is for DMA access)
 * will dealloc the sg table itself, via the sg_dma_access handle.
 *  [remember, sg_handle is "regular" allocation handle.
 *   sg_dma_access is for DMA access to the table]
 * 
*/

void dealloc_sg_table(struct sgclump *sg_clump){
	/* This supposedly frees both the 'access handle', AND
	 * the actual memory in the table
	 */
	DLPRINT2(1,"dealloc_sg_table(table@%p)",sg_clump->tableptr);

	if(sg_clump->sg_dma_access!=NULL){
		cleanup_dma_mem(&(sg_clump->sg_handle),
				&(sg_clump->sg_dma_access));
	}
}
/* realloc_sg_table:
 * Nasty hack to hopefully work around bug in HPT366 hardware.
 * sometimes the DMA engine chokes on a particular request.
 * So change the address of the scatter/gather table by reallocating it.
 * We will free up the old table, after copying the info to the new one.
 * return 0 if okay, or 1 if something went wrong with alloc_sg_table
 */
int realloc_sg_table(struct hpt_softstate *softstate, struct sgclump *sg_clump){
	struct sgclump oldinfo;
	uint64_t blocknum;
	
	bcopy(sg_clump, &oldinfo, sizeof(struct sgclump));

	if(sg_clump->bufptr==NULL){
		blocknum=0;
	} else {
		blocknum=sg_clump->bufptr->b_lblkno;
	}
	DLPRINT3(0, "REALLOCATING sg_table at %p (block num %llx)\n",
			sg_clump->tableptr, blocknum);

	if(alloc_sg_table(softstate,sg_clump) !=0){
		return 1;
	}
	/* XXX Not entirely sure this will work: we should probably
	 * use a ddi_mem_rep_put32 or something...?
	 */
	bcopy(&(oldinfo.tableptr),sg_clump->tableptr, SGTABLESIZE);

	ddi_dma_sync(sg_clump->sg_handle, 0,0, DDI_DMA_SYNC_FORDEV);

	dealloc_sg_table(&oldinfo);

	return 0;
	
}

/* DEBUG routine, for looking at the contents of a particular
 * SG table, on a particular sgclump struct.
 * Make really REALLY sure this sgclump is guaranteed not to
 * have any changes to it during the print.
 */
void printsgtable(struct sgclump *clumpptr){
	int entcount;
	struct sg_entry *sg_slot;
	ddi_acc_handle_t sg_access;
	uint32_t ioaddr,datalen;

	if(clumpptr==NULL){
		printf("printsgtable: ERROR: blank clumpptr");
		return;
	}
	printf("DEBUG: printsgtable on clumpptr %p",clumpptr);

	sg_slot=clumpptr->tableptr;
	sg_access = clumpptr->sg_dma_access;
	for(entcount=0;entcount<HPT_MAX_SG_ENT;entcount++){
		ioaddr=ddi_get32(sg_access, &sg_slot->ioaddr);
		datalen=ddi_get32(sg_access, &sg_slot->datalen);
		printf("  ioaddr=%x, datalen=%x",ioaddr,datalen);
		if(datalen & SCATTER_EOT){
		/*
			datalen &= ~SCATTER_EOT;
			printf("  EOT set. actual datalen==%x",
			       datalen);
		*/
			return;
		}
		sg_slot++;
	}

}

/*
 * Given a particular (struct buf*), and an existing sg_clump,
 * will attempt to ALLOCATE AND BIND a brand new dma_handle for thebuf,
 * and point sg_clump->bufhandle to it.
 *
 * return 0 on okay, 1 on fail.
 */
int create_and_bind(dev_info_t *dip, struct buf *bp,
		struct sgclump*sg_clump,int bindflags,
		ddi_dma_cookie_t *cookieptr,int *ccptr){

	int bindret;

	/* This is for dma-binding of the user buf to device memory */
	ddi_dma_attr_t tmp_dma_attr = buf_bind_attributes;

	if(ddi_dma_alloc_handle(dip,
				&tmp_dma_attr,
				DDI_DMA_SLEEP, NULL,
				&(sg_clump->bufhandle)) != DDI_SUCCESS) {
		cmn_err(CE_WARN,"hpt: could not allocate dma handle");
		return 1;
	}
	bindret=ddi_dma_buf_bind_handle(sg_clump->bufhandle,
		bp, bindflags, NULL,0,
		cookieptr, ccptr);

	if(bindret!=DDI_DMA_MAPPED) {
		cmn_err(CE_WARN,"hpt: could not bind dma handle");

		/* yes, sometimes this actually needs doing, for some
		 * stupid reason
		 */
		if(sg_clump->bufhandle!=NULL){
			ddi_dma_unbind_handle(sg_clump->bufhandle);
			ddi_dma_free_handle(&(sg_clump->bufhandle));
		}
		return 1;
	}

	sg_clump->bufptr=bp;

	DLPRINT(2,"create_and_bind() successfull!");

	if(bindflags& DDI_DMA_WRITE){
		/* I would think this wouldn't be neccessary. but it **IS***/
		ddi_dma_sync(sg_clump->bufhandle,0,0,DDI_DMA_SYNC_FORDEV);
	}
	return 0;
}



/*************************************************************
 * fill_sgtable
 * NOTE: Radically chained from prior versions.
 *
 * NOTE2: radically changed again. Sigh.
 *
 * SUMMARY
 *  Looked at the passed in buf chain.
 * (chained via buf-av_forw)
 *
 * Allocates an sgclump.
 *  links the first bufptr to it
 *  binds all the appropriate handles in sgclump for it
 *
 * If there are additional bufptrs, allocates a new sgclump,
 * and recursively does binds on a one-to-one basis.
 *
 * but only the initial sgclump will get an SGTABLE allocated for the
 * whole request set.
 *
 * returns a pointer to the head sgclump allocated.
 *
 *  It is up to the caller to then move the stuff over to the active queue.
 *  It is also up to the caller to ENSURE that jobs in the passed in chain
 *  are all contiguous requests , that are either all READ, or all WRITE.
 *
 *  - do a ddi_dma_sync() for the entire table once entries for
 *    each DMA-bound buf have been added.
 *
 *
 * If something goes horribly wrong, may return NULL.
 *
 * WARNINGS:
 *  Caller must ensure chain total bytes does not exceed IDEMAXTRANSFER
 *
 */
struct sgclump* fill_sgtable(struct hpt_softstate *softstate,int drive,
			     struct buf*firstbuf ){
	dev_info_t *dip=softstate->dip;
	
	struct sg_entry *sg_freeslot=NULL;
	ddi_dma_handle_t sg_handle=NULL;
	ddi_acc_handle_t sg_dma_access=NULL;
	uint32_t ent_count=0;
	uint32_t bytecount=0; /* tmp accumilator */
	int readflag, bindflags;
	struct buf *newrequest;
	struct sgclump *headclump=NULL,*newclump=NULL;

	DLPRINT(2,"Entered fill_sgtable");

	newrequest=firstbuf;
	if(newrequest==NULL){
		printf("ERROR: fill_sgtable called with NULL?");
		return NULL;
	}


	readflag=newrequest->b_flags&B_READ;
	if(readflag){
		bindflags=DDI_DMA_READ;
	} else {
		bindflags=DDI_DMA_WRITE;
	}

	while(newrequest!=NULL){
		ddi_dma_cookie_t cookie;
		uint_t ccount;
		int errflag=0;
		struct sgclump* tmpsgclump;

		tmpsgclump=kmem_zalloc(sizeof(struct sgclump),KM_SLEEP);
		if(headclump==NULL){
			DLPRINT(2,"  alloc_sg_table: Allocating sgtable for bufchain");
			headclump=tmpsgclump;
			
			if(alloc_sg_table(softstate,headclump)!=0)
			{
				printf("ERROR: call to alloc_sg_table failed");
				kmem_free(headclump,sizeof(struct sgclump));
				return NULL;
			}
			sg_freeslot=headclump->tableptr;
			sg_handle=headclump->sg_handle;
			sg_dma_access=headclump->sg_dma_access;
			headclump->retry_count=num_retries;

		} else {
			/* set OLD clump to chain to new clump */
			newclump->next=tmpsgclump;
		}
		newclump=tmpsgclump;
		newclump->next=NULL;

		/* A handle for access to the sg-TABLE is already allocated.
		 * Now get a handle for this specific data transfer area.
		 *  dma handle will get bound to sgclump->bufhandle
		 */
		if(create_and_bind(dip, newrequest,newclump,bindflags,
				    &cookie,&ccount)!=0){
			errflag=1;
		}
		if(ccount+ ent_count>=SGNUMENTS){
			errflag=1;
		}

		/* If any errs, have to nullify ENTIRE DAMN THING */
		if(errflag==1){
			printf("ALLOC_SG_TABLE: ERR HIT!!");
			dealloc_sg_table(headclump);
			while(headclump!=NULL){
				if(headclump->bufhandle!=NULL){
					ddi_dma_unbind_handle(headclump->bufhandle);
					ddi_dma_free_handle(&(headclump->bufhandle));
				}

				newclump=headclump->next;
				kmem_free(headclump,sizeof(struct sgclump));
				headclump=newclump;
			}

			return NULL;
		}


		/* keep track to make sure we somehow dont go over max
		 * ATA transfer
		 */
		ent_count+=ccount;

		if(sg_freeslot==NULL){
			printf("CANNOTHAPPEN: fill_sgtable with NULL sg_freeslot\n");
			return NULL;
		}
		if(sg_handle==NULL){
			printf("CANNOTHAPPEN: fill_sgtable with null sg_handle\n");
			return NULL;
		}
		if(sg_dma_access==NULL){
			printf("CANNOTHAPPEN: fill_sgtable with null sg_dma_access\n");
			return NULL;
		}

		/* Program table with addresses of alloc'd DMA chunks now */
		while(ccount-- >0){
			ddi_put32(sg_dma_access,  &sg_freeslot->ioaddr,
				  cookie.dmac_address);


			DLPRINT3(2,"  PHYSaddr 0x%x(size%lx) added to sgtable",
				 cookie.dmac_address,cookie.dmac_size);


			bytecount=cookie.dmac_size;
			ddi_put32(sg_dma_access,  &sg_freeslot->datalen,
				  bytecount);

			if(ccount>0){
				ddi_dma_nextcookie(newclump->bufhandle,&cookie);
				sg_freeslot++;
			}
		}



		newrequest=newrequest->av_forw;

		if(newrequest!=NULL){
			/* Okay, lets take another turn around! */
			sg_freeslot++;
		}

	} /* while(newrequest!=NULL) */

	/* NOTE: sg_freeslot must point to last entry we made in sg table!*/
	 
	bytecount |= SCATTER_EOT;
	ddi_put32(sg_dma_access,  &sg_freeslot->datalen,  bytecount);
		
	ddi_dma_sync(sg_handle, 0,0, DDI_DMA_SYNC_FORDEV);

	
	return headclump;

}/* end of fill_sgtable */



/* This assumes we have a FREE-FLOATING (struct buf *), that has
 * already been detached from any waiting or active queue.
 * Go free up sg_clump, for each buf, plus any chained sg_table and handles.
 * Call biodone on each, and all children chained off bptr->av_forw
 * It will also call bioerror() IF and ONLY if "err" != 0.
 * in fact, it will set err to be what you specify as the err value
 */
void cleanup_iobuf(struct buf *bptr, int err){
	/* Now unbind all the DMA resources for each buf that
	 * just got serviced.
	 * The unbind on bufhandle does an implicit ddi_dma_sync(), supposedly
	 */
	struct buf *nextptr=NULL;

	while(bptr!=NULL){
		struct sgclump *sg_clump;
		sg_clump=(struct sgclump*)bptr->b_private;
		bptr->b_private=NULL;

		bptr->b_resid=0;


		if(bptr->b_flags & B_READ){
			DLPRINT3(1,"cleanup_iobuf calling biodone on buf %p (read block %lld)",
			 bptr,	bptr->b_lblkno);
		} else {
			DLPRINT3(1,"cleanup_iobuf calling biodone on buf %p (WRITE block %lld)",
			 bptr,	bptr->b_lblkno);
		}

		if(sg_clump!=NULL){
			/* clean up extra sg table memory */
			/* In theory, this should only be for the FIRST ENTRY*/
			/* We need to do this BEFORE biodone called
			 * on the buf
			 */
			if(nextptr!=NULL){
				printf("**WARNING**: sg on non-first");
			}
			if(sg_clump->bufhandle!=NULL){
				ddi_dma_unbind_handle(sg_clump->bufhandle);
				ddi_dma_free_handle(&(sg_clump->bufhandle));
			}
			
			dealloc_sg_table(sg_clump);

			kmem_free(sg_clump,sizeof(struct sgclump));
		}
		
		
		nextptr=bptr->av_forw;

		/* I dont know wny, but if we call bioerror higher up,
		 * the bptr->b_flags gets set to ZERO SOMEWHERE??
		 */
		if(err!=0){
			DLPRINT2(0,
				 "hpt:cleanup_iobuf calling bioerror(%d)\n",
				 err);
			bptr->b_resid=bptr->b_bcount;
			bioerror(bptr,err);
			if(bptr->b_flags &B_ERROR){
			DLPRINT(1,"DEBUG: after bioerr, B_ERROR flag is set\n");
			
			} else {
			DLPRINT(1,"DEBUG: after bioerr, B_ERROR flag is NOT set\n");
			}
		}
		biodone(bptr);
		bptr=nextptr;
	}
}



/* install_sgtable
 *
 * Looks up the sg_table in the "active" queue for the disk
 * (which must already be COMPLETELY FILLED OUT!!!)
 *  and lets the UDMA hardware know where the table is, in DMA memory.
 *
 * According to intel spec, you should call this BEFORE doing the
 * IDE DMA command
 *
 * We do an extra ddi_dma_sync() on the sg_table here, just to be paranoid
 *
 * return 1 on fail, 0 on okay
 *
 * It should only fail if you give an unsupported disk number,
 * or nothing on active queue (which is considered an error)
 */
int install_sgtable(struct hpt_softstate * softstate, int disk){
	uint32_t *UDMAsgptr;
	struct sgclump*sg;
	int bus=0;
	if(disk>1)bus=1;
	
	sg=hpt_getactivesgclump(softstate,bus);
	if(sg==NULL){
		printf("ERROR: install_sgtable found NULL on activequeue");
		return 1;
	}

	switch(disk){
	    case 0:
		UDMAsgptr=&(softstate->hpt_udmaregs->scat_gath1);
		break;
	    case 1:
		UDMAsgptr=&(softstate->hpt_udmaregs->scat_gath2);
		break;
	    default:
		return 1;
	}

	if(ddi_dma_sync(sg->sg_handle, 0,0, DDI_DMA_SYNC_FORDEV)!=DDI_SUCCESS){
		DLPRINT(0,"ERROR with ddi_dma_sync in install_sgtable?!!\n");
		/* should never happen. ha ha */
	}
	ddi_put32(softstate->hpt_udmaregs_handle, UDMAsgptr, sg->physaddr);

	return 0;
}


/**********************************************************************
 *    End of sg table section                                         *
 **********************************************************************/


/* The UDMA engine status register is separate from the IDE status
 * register used in printstatus().
 * The odd thing is, I'm not sure whats up with
 * "Primary Channel/Secondary Channel", since there's also
 *  "Drive 0/Drive 1".
 *
 * Anyway, look at printudmastatus() to decode what the bits mean.
 */
uint8_t readudmastatus(struct hpt_softstate *softstate,int bus){
	uint8_t *UDMAstatusptr, reg8;

	if(bus==0){
		UDMAstatusptr=&softstate->hpt_udmaregs->status1;
	} else {
		UDMAstatusptr=&softstate->hpt_udmaregs->status2;
	}
	
	reg8=ddi_get8(softstate->hpt_udmaregs_handle, UDMAstatusptr);
	return reg8;

}


/* convenience routine to check if a particular
 * bus (?) has generated an interrupt
 *
 * Except this is only tested for bus=0 right now
 *
 * return 1 if yes, 0 if no.
 */
int got_udma_intr(struct hpt_softstate * softstate, int bus){
	int reg8;

	reg8=readudmastatus(softstate,bus);
	if(reg8 & UDMASTATUS_IRQ){
		return 1;
	}

	return 0;

}

/* Checks for DMA error bit in UDMA status register.
 * return 1 on error-happened, 0 on clear 
 */
int got_udma_err(struct hpt_softstate * softstate, int bus){
	int reg8=readudmastatus(softstate,bus);
	if(reg8 & UDMASTATUS_ERR){
		return 1;
	}
	return 0;
}

/* Checks for BUSY flag in UDMA status register.
 * return 1 on busy, 0 on clear
 * (Okay, badly named, but it fits in with existing naming scheme)
 */
int got_udma_busy(struct hpt_softstate * softstate, int bus){
	int reg8=readudmastatus(softstate,bus);
	if(reg8 & UDMASTATUS_BUSY){
		return 1;
	}
	return 0;
}

/* clear IRQ status bit for UDMA hardware */
void clear_udma_intr(struct hpt_softstate * softstate, int bus){
	uint8_t *UDMAstatusptr, reg8;

	if(bus==0){
		UDMAstatusptr=&softstate->hpt_udmaregs->status1;
	} else {
		UDMAstatusptr=&softstate->hpt_udmaregs->status2;
	}

	reg8=UDMASTATUS_IRQ;

	/* status bits are cleared by writing 1 to them */
	ddi_put8(softstate->hpt_udmaregs_handle, UDMAstatusptr,reg8);

}

/* clear ERR status bit for UDMA hardware */
void clear_udma_err(struct hpt_softstate * softstate, int bus){
	uint8_t *UDMAstatusptr, reg8;

	if(bus==0){
		UDMAstatusptr=&softstate->hpt_udmaregs->status1;
	} else {
		UDMAstatusptr=&softstate->hpt_udmaregs->status2;
	}

	reg8=UDMASTATUS_ERR;

	/* status bits are cleared by writing 1 to them */
	ddi_put8(softstate->hpt_udmaregs_handle, UDMAstatusptr,reg8);

}

/* wait on UDMA status reg, while bus master is active,
 * aka "busy"
 * return 0 on done, 1 on error/timeout
 */
int UDMAwwbusy(struct hpt_softstate *softstate,int bus){
	int count;
	uint8_t reg8;
	uint8_t *UDMAstatusptr;


	switch(bus){
	    case 0:
		UDMAstatusptr=&softstate->hpt_udmaregs->status1;
		break;
	    case 1:
		UDMAstatusptr=&softstate->hpt_udmaregs->status2;
		break;
	    default:
		cmn_err(CE_WARN,"bad bus val in UDMAwwbusy");
		return 1;
	}

	for(count=0;count<100;count++){
		reg8=ddi_get8(softstate->hpt_udmaregs_handle, UDMAstatusptr);
		if(reg8 & UDMASTATUS_ERR){
			cmn_err(CE_WARN,"UDMAWwbusy: ERR bit set");
			return 1;
		}
		if((reg8 & UDMASTATUS_BUSY)==0){
			return 0;
		}
		drv_usecwait(1); /* used to be delay(1) */
	}

	cmn_err(CE_WARN,"UDMAwwbusy: timeout");
	return 1;
}



/* This is an intel-specific "enable UDMA/33 support" tweak,
 * for a specific drive, 
 * in the intel PIIX4 ide/udma controller chipset.
 * The specific offset 0x48 is NOT MENTIONED in the hpt documentation.
 */
void intel_udma_enable(struct hpt_softstate *softstate, int drive){
	uint8_t reg8,tmp8;
	int INTEL_UDMACONF=0x48;
	

	if(drive>3){
		printf("intel_udma_enable: bad drive num");
		return;
	}

	reg8=pci_config_get8(softstate->pci_conf_h, INTEL_UDMACONF);

	tmp8=(0x01 << drive);
	reg8=reg8 | tmp8;

	pci_config_put8(softstate->pci_conf_h, INTEL_UDMACONF,reg8);

}

/* sets some flags we like for the control registers appropriately.
 * This sets values for BOTH "channels"/buses
 * Should only call at attach time, methinks.
 */
void hpt_set_controls(struct hpt_softstate *softstate){
	uint16_t controloffset=0x50;
	uint16_t regval;

	
	regval=pci_config_get16(softstate->pci_conf_h,controloffset);
	DLPRINT2(0,"hpt_set_controls: initial value of pcicontrol@50=0x%x\n",regval);
	/* set "dont trigger interrupt until buffer clear */
	regval&=0xffff03;
	/* from high bits to low bits:
	 * NO "fast interrupt"
	 * Hold interrupt until buffer is empty
	 * Clear SG state after interrupt
	 * flush FIFO for both channels
	 * enable channels 0 and 1
	 * clear buffers
	 */
	regval|=0x7f3c;

	pci_config_put16(softstate->pci_conf_h,controloffset,regval);

	regval=0;/* paranoia, and tiny delay also maybe */

	regval=pci_config_get16(softstate->pci_conf_h,controloffset);
	DLPRINT2(1," after pcicontrolreg set, pcicontrol@50=0x%x\n",regval);
}

/*
 * Set the special controller timing values for a bus, based on
 * UDMA speed you are using
 */
void hpt_set_timings(struct hpt_softstate *softstate,int bus, int UDMAspeed){
	uint32_t timingval;
	uint8_t busspeed=0;
	uint32_t busspeed_offset=0x41 + bus*4;
	uint32_t timereg_offset=0x40 + bus*4;

	busspeed=pci_config_get8(softstate->pci_conf_h, busspeed_offset);

	/* XXX the hpt370 is less speed-sensitive.
	 * Unfortunately, this is currently for the hpt366
	 */
	if(busspeed==0x85){
		DLPRINT(1,"hpt_set_timings: bus speed==25mhz");
		switch(UDMAspeed){
			case MODE_UDMA2:
				timingval=0x90cf8521;
				break;
			case MODE_UDMA4:
				timingval=0x90cf8521;
				break;
			default:
				DLPRINT(0,"warning: using default timings");
				timingval=0x90cf8521;
				break;
		}

	} else if(busspeed==0xd9){
		DLPRINT(1,"hpt_set_timings: bus speed==40mhz");
		switch(UDMAspeed){
			case MODE_UDMA2:
				timingval=0x900bd943;
				break;
			case MODE_UDMA4:
				timingval=0x900fd943;
				break;
			default:
				DLPRINT(0,"warning: using default timings");
				timingval=0x0120d9d9;
				break;
		}

	} else {
		/* Pretty much all systems should really be here.
		 * 33mhz is standard for PCI bus speed, after all
		 * (But what do I use for 66mhz?)
		 */
		if(busspeed==0xa7){
			DLPRINT(1,"bus speed=33mz")
		} else {
			DLPRINT(1,"bus speed=DEFAULT??")
		}
		switch(UDMAspeed){
			case MODE_UDMA2:
				DLPRINT(1,"setting timings for UDMA2");
				timingval=0x90caa731;
				break;
			case MODE_UDMA4:
				DLPRINT(1,"setting timings for UDMA4");
				timingval=0x90c9a731;
				break;
			default:
				DLPRINT(1,"warning: using default timings");
				timingval=0x0120a7a7;
				break;
		}
		
	}

	/* both the BSD and linux drivers unset this stupid bit.
	 * So why the hell did they set it in the FIRST PLACE??
	 */
	/*timing&= ~0x80000000;*/
	pci_config_put32(softstate->pci_conf_h, timereg_offset,timingval);

}

int hpt_udma_getremaining(struct hpt_softstate * softstate, int bus){
	int reg_offset=0x50;
	uint32_t retval=0;
	if(bus==1) {
		reg_offset=0x54;
			printf("DEBUG: warning: getremaining called with bus=1\n");
	}
	
	retval=pci_config_get32(softstate->pci_conf_h, reg_offset);
	DLPRINT3(2,"DEBUG: getremaining raw32@0x%x==0x%x\n",reg_offset,retval);

	retval=(retval>>16);
	retval&=0x1ff;
	return retval;
	
}


/* Program the device on the specified channel, to use highest
 * UDMA mode available, up to UDMA4, using the SET_FEATURES ATA command.
 *
 * If we had a pure intel chipset, we would have to do special
 *  intel_udma_enable call here as well.
 *
 * MUTEXES
 *  It is assumed you have the appropriate bus mutex for the drive
 *
 * return 0 on okay, 1 on fail.
 */
int init_udma(struct hpt_softstate *softstate, int drive){
	uint8_t udma_mode=0;
	uint8_t feature=IDE_FEAT_TRANSMODE;
	uint8_t command=IDECMD_SETFEATURES;
	uint8_t feat_enabled;
	int bus=drive>>1;

	DLPRINT(2,"(udma_init Skipping call to intel_udma_enable())");

	if(drive>HPT_NUMDRIVES){
		cmn_err(CE_WARN,"Error: drivenum too high in progdma");
		return 1;
	}


	feat_enabled=softstate->dev_info[bus][88] & 0xff;

	if(feat_enabled==0){
		DLPRINT2(1,"hpt: drive %d does not support UDMA!",drive);
		return 1;
	}

	if(feat_enabled &0x01) udma_mode=MODE_UDMA0;
	if((feat_enabled &0x02) && (udma_max>=1)) udma_mode=MODE_UDMA1;
	if((feat_enabled &0x04) && (udma_max>=2)) udma_mode=MODE_UDMA2;
	if((feat_enabled &0x08) && (udma_max>=3)) udma_mode=MODE_UDMA3;
	if((feat_enabled &0x10) && (udma_max>=4)) udma_mode=MODE_UDMA4;

	if(ide_select_drive(softstate,drive)!=0){
		printf("ERR: init_udma could not select drive");
		printstatus(readaltstatus(softstate,bus));
		return 1;
	}

	/* This sets the parameter for the "sub-command", below */
	ddi_put8(softstate->hpt_ide_command_handle[bus],
		 &softstate->ide_command_regs[bus]->sector_count, udma_mode);

	if(wait_while_busy(softstate,bus)!=0){
		printf("ERROR: init_udma busied out");
		return 1;
	}

	/* This is actually the "features" register.
	 * read="error", write="feature"
	 * we write the "subcommand" of IDE_SETFEATURES here.
	 */
	ddi_put8(softstate->hpt_ide_command_handle[bus],
		 &softstate->ide_command_regs[bus]->error, feature);

	if(wait_while_busy(softstate,bus)!=0){
		printf("ERROR: init_udma busied out");
		return 1;
	}

	if(idecmd_checkstat(softstate,bus,command) != 0){
		printf("ERROR: init_udma could not write SETFEAT command");
		if(debug_level>0){
			printstatus(readaltstatus(softstate,bus));
		}
		return 1;
	}

	DLPRINT2(1,"hpt: init_udma enabled UDMA OK on drive %d",drive);

	softstate->drive_speed[drive]=udma_mode;

	/* why TWO flags? dunno. so be paranoid and check both*/
	if((softstate->dev_info[drive][83] & 0x02) != 0x02){
		cmn_err(CE_WARN,"queueing not okay for drive(feat-83)");
		softstate->queued_ok[bus]=0;
	}
	if((softstate->dev_info[drive][86] & 0x02) != 0x02){
		cmn_err(CE_WARN,"queueing not okay for drive(feat-86)");
		softstate->queued_ok[bus]=0;
	}

	return 0;
}



/*
 * This routine handles programming and start of UDMA chipset.
 *  You also have to program the IDE chipset, FIRST, for a UDMA transfer
 *  to actually happen
 *
 * This ASSUMES
 *    1. the actual IDE-level command has been sent to the IDE chips
 *    2. The sg table clump address has already been programmed into
 *	 the UDMA hardware
 *
 * We then set the UDMA command register, with the read/write bit set
 * as flagged by the readflag argument.
 *
 * Return 0 for okay, 1 on fail.
 *   But since we dont check status, its almost impossible to fail, unless
 *   you give the wrong arg or something.
 *
 * MUTEXES:
 *	You should probably aquire the bus_mutex before calling this.
 *	No mutexes are changed in here.
 */
int start_udma(struct hpt_softstate * softstate, int disk, int readflag){

	uint8_t *UDMAcommandptr;
	uint8_t reg8;
	int bus=0;
	if(disk>1) bus=1;

	switch(disk){
	    case 0:
		UDMAcommandptr=&softstate->hpt_udmaregs->command1;
		break;
	    default:
		return 1;
	}
	if(readflag){
		DLPRINT(1,"start_udma setting READ flag");
		reg8=UDMACMD_DISKREAD;/* set UDMA DiskRead flag */
	} else {
		DLPRINT(1,"start_udma NOT setting read flag");
		reg8=0;
	}

	ddi_put8(softstate->hpt_udmaregs_handle,UDMAcommandptr,reg8);
	reg8=readaltstatus(softstate,bus); /* delay */

	if((readflag==0) && (debug_level>0)){
		uint32_t reg32;
		
		DLPRINT(1,"DEBUG: start_udma before start of a WRITE call");

		reg32=ddi_get32(softstate->hpt_udmaregs_handle,
				&(softstate->hpt_udmaregs->scat_gath1));
		DLPRINT2(1," Scatter/Gather tablereg points to %x\n",
		       reg32);
		       
	}

	reg8=ddi_get8(softstate->hpt_udmaregs_handle,UDMAcommandptr);
	reg8 |= UDMACMD_START;	/* "Start DMA engine"! */
	ddi_put8(softstate->hpt_udmaregs_handle,UDMAcommandptr,reg8);

#ifdef BYHAND_DETECT /* DONT DO THIS!! */
	if(got_udma_err(softstate,bus)){
		printf("start_udma: EARLY ERROR DETECT");
	}
#endif

	return 0;
}

/*
 * clear "START DMA operation" bit in UDMA hardware
 *
 * This routine is both for "good" and "bad" things:
 *  "good" == A DMA operation has completed. You still have to turn off
 *            the DMA engine, though.
 *
 *  "bad"  == If you have some kind of error condition, and you need
 *            to stop any UDMA operation in progress.
 *            (like before resetting a drive)
 *
 */
void stop_udma(struct hpt_softstate * softstate, int bus){
	uint8_t *UDMAcommandptr, reg8;

	if(bus!=0){
		printf(" hpt -- ERROR: stop_udma cannot handle bus>0");
		return;
	}

	UDMAcommandptr=&softstate->hpt_udmaregs->command1;
	reg8=ddi_get8(softstate->hpt_udmaregs_handle, UDMAcommandptr);

	reg8 &= ~UDMACMD_START;
	ddi_put8(softstate->hpt_udmaregs_handle, UDMAcommandptr,reg8);

	/* just a timing thing. I hope. But be paranoid and check
	 * the actual value
	 */
	reg8=ddi_get8(softstate->hpt_udmaregs_handle, UDMAcommandptr);
	if(reg8& UDMACMD_START){
		printf(" hpt: ERROR: stop_udma did not stop engine!");
	}
	
	
}

/* YOU MUST HAVE THE BUS MUTEX BEFORE CALLING THIS.
 * We do not fiddle with mutexes or semaphores.
 * This is a shared subroutine for start_one_job, and
 * hpt_handle_intr
 *
 * It assumes there is a job on the active queue for the drive,
 * and you want the hardware to start doing something with it!
 *
 * return 0 on okay, 1 on bad
 */
int hpt_start_io(struct hpt_softstate *softstate, int drive)
{
	uint8_t idecmd;
	int readflag,numblocks;
	struct buf *bufptr;
	int bus=0,minornum,instance;

	if(drive>1) bus=1;


	DLPRINT(1,"in hpt_start_io");

	numblocks=hpt_countactiveblocks(softstate,drive);
	if(numblocks==-1){
		printf("ERROR: hpt_countactiveblocks returned -1\n");
		return 1;
	}

	bufptr=hpt_getactivehead(softstate,bus);

	if(bufptr==NULL){
		printf("hpt_start_io: activequeue==NULL??");
		return 1;
	}

	minornum=getminor(bufptr->b_edev);
	instance=GETINSTANCE(minornum);
	
	if((bufptr->b_flags &B_READ) !=0){
		idecmd=IDECMD_READDMA;
		readflag=1;
	} else {
		idecmd=IDECMD_WRITEDMA;
		readflag=0;
	}

	DLPRINT2(1,"Doing the IDE hardware positioning to raw blknum 0x%llx",
		bufptr->b_lblkno);

	positiondisk(softstate, drive,
		     bufptr->b_lblkno,
		     (numblocks&0xff));

	DLPRINT(2,"Installing sgtable into UDMA register");

	install_sgtable(softstate,drive);

	/* Now write the actual IDE COMMAND to the IDE hardware.
	 * Note that this is step 1 of 2, to actually start things.
	 */

	if(idecmd_nocheck(softstate,bus,idecmd) != 0){
		printf("ERROR(hpt): hpt_start_io: idecmd failed!");
		return 1;
	} else {
		DLPRINT(1,"hpt_start_io:  idecmd_nocheck() returned OK");
	}

#ifndef SOFTINTR_TIMEOUT /* If not "faking" timeout with softintr...*/

	/* Have to do this BEFORE set up of hardware, otherwise we can
	 * trigger interrupt before we start the timer!
	 * Plus, keep this inside mutex, to reduce possibly
	 * overriding another timer.
	 */
	if(softstate->timeout_id[bus]!=0){
		DLPRINT2(1," removing old timeout %d before using new one",
			(int)softstate->timeout_id[bus]);
		untimeout(softstate->timeout_id[bus]);
		softstate->timeout_id[bus]=0;
	}
	DLPRINT(2,"Setting up 15-sec timeout()");
	/* make a 10 second timeout for this I/O */
	softstate->timeout_id[bus]=timeout(hpt_timeout_handler,
					   (void*)((instance<<8)|drive),
					15*drv_usectohz(1000000));
	DLPRINT2(1,"set up timeout id %d",(int)softstate->timeout_id[bus]);
#endif

	/* This starts the "UDMA Engine", after the IDE side 
	 * has already started.
	 * 
	 * Keep in mind that AS SOON AS we trigger this, the interrupt
	 * could stop us in our tracks, and release the semaphore
	 */
	if(start_udma(softstate,drive,readflag)!=0){
		DLPRINT(0,"ERROR: start_udma returned non-0");
		return 1;
	}

	DLPRINT(1,"DEBUG: hpt_startio: start_udma returned ok. Done.\n");
	/* Note: MUST NOT READ IDE REGISTERS NOW TO CHECK, OR CAN HANG THINGS*/
	return 0;
}

/* start_one_job: a simplified routine for strategy to call,
 * to queue up a data request and start I/O on it.
 * Despite the name, it can handle read OR write requests.
 * Supposedly. I'm having problems with consistent writes, currently.
 *
 * This function start as a test module, but it is slowly becoming a
 * production one.
 *
 * This assumes that there is a single job already in our waitqueue for
 * the particular drive.
 *
 * THis will move a job to the active queue slot, and trigger the
 * operation (which will generate an interrupt when done, OR err flagged)
 * Then it's all up to the interrupt handler
 *
 * return 0 on okay, 1 on fail.
 *
 * We handle calling bioerror()/biofail(), if it is neccessary,
 * so there isnt really anything the caller can do if we return 1.
 *
 * MUTEXES:
 *	We call the waitqueue handling routines, which will presumably use
 *	the waitqueue_mutex[]
 *	We personally take and then release bus_mutex[bus]
 *	We also call the activequeue handling routines, which will
 *	presumably decrement the activequeue_semaphore
 *
 */
int hpt_start_one_job(struct hpt_softstate *softstate, int drive)
{
	int bus=0;
	struct buf *bufptr;
	struct sgclump*clumpptr;
	int retval=0; /* 0==normal exit */
	if(drive>1) bus=1;

	/*
	 * INTEL 440BX IDE spec says:
	 * - give sgtable physaddr to udma
	 * - send STARTDMA idecmd to disk
	 * - turn on START bit to udma commandreg
	 * - (interrupt is triggered)
	 * -   reset START bit in udma commandreg, in interrupt handler
	 */
	DLPRINT(2,"start_one_job entered");


	bufptr=hpt_one_from_waitqueue(softstate,drive);

	if(bufptr==NULL){
		printf("ERROR: waitqueue is EMPTY??");
		return 1;
	}

	DLPRINT(2,"  start_one_job: got queued job, going for fill_sgtable");

	/* THis is expected to allocate an sg_clump + sg table, and
	 * return the sg_clump* on bufptr->b_private
	 */

	
	clumpptr=fill_sgtable(softstate,drive, bufptr);

	if(clumpptr==NULL){
		printf("ERROR: hpt_start_one_job: fill_sgtable Erred");
		return 1;
	}


	DLPRINT(2,"  start_one_job: Done fill_sgtable!");
	DLPRINT2(1,"  start_one_job: making block %lld active now",bufptr->b_lblkno);


	/* We're DONE with queue fiddlings. Now we just have to trigger
	 * the actual hardware operation!
	 */

	DLPRINT(1,"start_one_job: going for set_activequeue");

	/* This calls sema_p on the activequeue semaphore */
	if(hpt_set_activequeue(softstate,bus,clumpptr)!=0){
		printf("start_one_job: something already on active queue??");
		printf("    erroring out IO job");
		hpt_free_sgclumpchain(clumpptr,EIO);
		return 1;
	}

	DLPRINT(1,"start_one_job: done activequeue, going for busmutex");
	mutex_enter(&softstate->bus_mutex[bus]);

	/* Now set up disk, and trigger DMA operation */
	DLPRINT(1,"start_one_job going for start_io");
	if(hpt_start_io(softstate,drive)!=0){
		printf("ERROR: start_one_job failed to start_io. Calling cleanup_active.");
		if(hpt_cleanup_active(softstate,drive,EIO)==0){
			DLPRINT(1,"   (released semaphore on activequeue).");
			sema_v(&softstate->activequeue_sema[bus]);
		}
		retval=1;
	}

	
	mutex_exit(&softstate->bus_mutex[bus]);
	DLPRINT2(1,"start_one_job: released bus_mutex and returning %d",
		retval);

	/* Do NOT free up softstate->queue_sema[bus], because
	 * interrupt handler will do that when job is actually done
	 */

	/* According to ATA spec, we're supposed to get an interrupt
	 * on success OR FAILURE of the command.
	 * So it's best I dont try reading the registers now
	 */

	
	return retval;
}


/*
 * mainly, called by hpt_handle_intr.
 * for now, throw away result. this is just to hopefully reset DMA.
 * Yes, I know this is hellaciously long for an interrupt routine
 * to call :-(
 * Shouldnt be bad on dual processor systems, but uniprocessor
 * systems might get a little unhappy :-/
 *
 * MUTEXES:
 *   You must hold the bus mutex before calling this. We do not touch it.
 *
 * We explicitly disable, then enable, interrupts on the bus.
 */
void hpt_read_dma_polled(struct hpt_softstate *softstate, int drive,
		    uint64_t blockno)
{
	dev_info_t *dip = softstate->dip;
	struct sg_entry *sg_slot=NULL;
	ddi_dma_handle_t buffer_handle,sg_handle;
	ddi_dma_attr_t tmp_dma_attr;
	ddi_acc_handle_t buffer_acc_handle,sg_acc_handle;
	size_t buffer_padded_len,sg_padded_len;
	caddr_t buffer_addr,sg_addr;
	ddi_dma_cookie_t buffer_cookie,sg_cookie;
	int buffer_ccount,sg_ccount, funcstat, loopcount;
	uint8_t idecmd=IDECMD_READDMA;
	int bus=0;
	if(drive>1) bus=1;

	tmp_dma_attr = sg_attributes;

	/* First, do transfer buffer allocation, etc*/
	if(ddi_dma_alloc_handle(dip, &tmp_dma_attr,
	    DDI_DMA_DONTWAIT, NULL, &buffer_handle)
	    != DDI_SUCCESS) {
		printf("hpt_read_dma_polled: could not alloc handle");
		return;
	}
	if(ddi_dma_mem_alloc(buffer_handle, 512, &dev_attr,
			     DDI_DMA_CONSISTENT, DDI_DMA_DONTWAIT, NULL,
			     &buffer_addr, &buffer_padded_len,
			     &buffer_acc_handle) != DDI_SUCCESS)
	{
		printf("hpt_read_dma_polled: could not alloc bufmem");
		ddi_dma_free_handle(&buffer_handle);
		return;
	}
	/* Note that yes, we have to pass PADDED_LEN, according to
	 * ddi_dma_mem_alloc() manpage
	 */
	funcstat=ddi_dma_addr_bind_handle(buffer_handle, NULL,
			       buffer_addr, buffer_padded_len,
			       DDI_DMA_RDWR|DDI_DMA_CONSISTENT,
			       DDI_DMA_DONTWAIT, NULL,
			       &buffer_cookie, &buffer_ccount);
	if((funcstat!= DDI_DMA_MAPPED) || (buffer_ccount!=1)){
		DLPRINT(0,"bind_handle failed for buffer");
		ddi_dma_mem_free(&buffer_acc_handle);
		ddi_dma_free_handle(&buffer_handle);
		return;
		
	}

	/**** Now, allocate/bind our sg table ***/
	if(ddi_dma_alloc_handle(dip, &tmp_dma_attr,
	    DDI_DMA_DONTWAIT, NULL, &sg_handle)
	    != DDI_SUCCESS) {
		DLPRINT(0,"hpt_read_dma_polled: could not alloc sghandle");
		cleanup_dma_mem(&buffer_handle, &buffer_acc_handle);
		return;
	}
	if(ddi_dma_mem_alloc(sg_handle, 512, &dev_attr,
			     DDI_DMA_CONSISTENT, DDI_DMA_DONTWAIT, NULL,
			     &sg_addr, &sg_padded_len,
			     &sg_acc_handle) !=DDI_SUCCESS)
	{
		printf("hpt_read_dma_polled: could not alloc sgmem");
		cleanup_dma_mem(&buffer_handle, &buffer_acc_handle);
		ddi_dma_free_handle(&sg_handle);
		return;
	}

	funcstat=ddi_dma_addr_bind_handle(sg_handle, NULL,
			       sg_addr, sg_padded_len,
			       DDI_DMA_RDWR|DDI_DMA_CONSISTENT,
			       DDI_DMA_DONTWAIT, NULL,
			       &sg_cookie, &sg_ccount);
	if((funcstat!= DDI_DMA_MAPPED) || (sg_ccount!=1)){
		printf("bind_handle failed for sg");

		cleanup_dma_mem(&buffer_handle, &buffer_acc_handle);

		ddi_dma_mem_free(&sg_acc_handle);
		ddi_dma_free_handle(&sg_handle);
		return;
		
	}


	/*************************************************************\
	 *    PHEW! We finally have the resources allocated. I hope. *
	\*************************************************************/
			       

	sg_slot=(struct sg_entry*)sg_addr;
	ddi_put32(sg_acc_handle, &sg_slot->ioaddr, sg_cookie.dmac_address);
	ddi_put32(sg_acc_handle, &sg_slot->datalen, 512 | SCATTER_EOT);

	if(ddi_dma_sync(sg_handle, 0,0, DDI_DMA_SYNC_FORDEV)!=DDI_SUCCESS){
		printf("COULD NOT SYNC sgtable??");
		cleanup_dma_mem(&buffer_handle, &buffer_acc_handle);
		cleanup_dma_mem(&sg_handle, &sg_acc_handle);
		return;
	}
	disable_interrupt(softstate,bus);

	if(positiondisk(softstate,drive,blockno,1) != 0){
		printf(" ERR: could not positiondisk for read_poll");
		cleanup_dma_mem(&buffer_handle, &buffer_acc_handle);
		cleanup_dma_mem(&sg_handle, &sg_acc_handle);

		enable_interrupt(softstate,bus);
		return;
	}
	
	/* equivalent to "install_sgtable" */
	switch(drive){
		case 0:
		ddi_put32(softstate->hpt_udmaregs_handle,
			  &softstate->hpt_udmaregs->scat_gath1,
			  sg_cookie.dmac_address);
		break;
		case 1:
		ddi_put32(softstate->hpt_udmaregs_handle,
			  &softstate->hpt_udmaregs->scat_gath2,
			  sg_cookie.dmac_address);
		break;
		default:
			printf("drive number too high for read_polled");
			cleanup_dma_mem(&buffer_handle, &buffer_acc_handle);
			cleanup_dma_mem(&sg_handle, &sg_acc_handle);

			enable_interrupt(softstate,bus);
			return;
		break;
	
	}

	if(idecmd_nocheck(softstate,bus,idecmd) != 0){
		printf("ERROR: read_dma_polled could not issue IDEcmd");
		cleanup_dma_mem(&buffer_handle, &buffer_acc_handle);
		cleanup_dma_mem(&sg_handle, &sg_acc_handle);

		enable_interrupt(softstate,bus);
		return;
	}
	if(start_udma(softstate,drive,1)!=0){
		printf("ERROR: read_dma_polled could not start_udma");
		cleanup_dma_mem(&buffer_handle, &buffer_acc_handle);
		cleanup_dma_mem(&sg_handle, &sg_acc_handle);

		enable_interrupt(softstate,bus);
		return;
	}

	DLPRINT(1,"Spinning on UDMA wait");
	loopcount=10000;
	while(got_udma_busy(softstate,bus)){
		drv_usecwait(10);
		loopcount--;
		if(loopcount<=0){
			printf("Err: UDMA Poll timed out");
			break;
		}
	}

	if(got_udma_err(softstate,bus)){
		printf( "ERROR: poll on udma showed UDMA ERR");
	}
	clear_udma_intr(softstate,bus);
	{
		uint8_t stat;
		stat=readstatus(softstate,bus);
		printstatus(stat);
	}
	

	DLPRINT(1, "UDMA via poll DONE. enabling interrupt again.");
	enable_interrupt(softstate,bus);

	/* ALL DONE. Clean up now */
	cleanup_dma_mem(&buffer_handle, &buffer_acc_handle);
	cleanup_dma_mem(&sg_handle, &sg_acc_handle);
	return;
}

