/* This file should be used by non"hpt-udma.c" files, for externs.
 * 
 * But it is ALSO used by hpt-udma.c
 *
 * @(#) hpt-udma.h 1.23@(#)
 */

int start_udma(struct hpt_softstate *, int drive,int readflag);

int install_sgtable(struct hpt_softstate *, int drive);

int alloc_sg_table(struct hpt_softstate *softstate, struct sgclump*);
int realloc_sg_table(struct hpt_softstate *softstate, struct sgclump*);
void dealloc_sg_table(struct sgclump *sg_clump);
struct sgclump *fill_sgtable(struct hpt_softstate *, int drive, struct buf* );
void printsgtable(struct sgclump*);

void cleanup_iobuf(struct buf *bptr, int errflag);
/* call cleanup_active only if driver severely screwed up */
void cleanup_active(struct hpt_softstate *softstate,int drive, int errflag);

int init_udma(struct hpt_softstate *softstate, int drive);
void hpt_set_timings(struct hpt_softstate *softstate,int bus, int UDMAspeed);
void hpt_set_controls(struct hpt_softstate *softstate);

void printudmastatus(uint8_t reg8);
uint8_t readudmastatus(struct hpt_softstate *softstate,int bus);

int got_udma_intr(struct hpt_softstate * softstate, int bus);
int got_udma_err(struct hpt_softstate * softstate, int bus);
int got_udma_busy(struct hpt_softstate * softstate, int bus);
void clear_udma_intr(struct hpt_softstate * softstate, int bus);
void clear_udma_err(struct hpt_softstate * softstate, int bus);
int hpt_udma_getremaining(struct hpt_softstate * softstate, int bus);

/* insert job into queue, then start queue processing */
int hpt_start_one_job(struct hpt_softstate *softstate,int drive);

/* start hardware DOING something with job on active queue*/
int hpt_start_io(struct hpt_softstate *softstate, int drive);

void stop_udma(struct hpt_softstate * softstate, int bus);

/* "wait while busy" */
int UDMAwwbusy(struct hpt_softstate *softstate,int bus);


/************************************************************/
/*      #defines and structs only, below this point         */
/************************************************************/

/* Maybe this bit belongs in hpt-ide.h */
/* MODE_* is for use with IDE_FEAT_TRANSMODE */
#define MODE_UDMA0		0x40
#define MODE_UDMA1		0x41
#define MODE_UDMA2		0x42
#define MODE_UDMA3		0x43 /* should check for 80-cable first*/
#define MODE_UDMA4		0x44 /* should check for 80-cable first*/



/* Scatter/Gather Table length, for DMA activities  */
#define SGENTSIZE 8			/* byte size of one entry */
#define HPT_MAX_SG_ENT 16		/* total allowed in table */
#define SGNUMENTS HPT_MAX_SG_ENT
#define SGTABLESIZE (SGENTSIZE * SGNUMENTS)
#define SCATTER_EOT 0x80000000	/* goes in "special" for last entry*/


/* An "Entry" in a scatter/gather table looks like this */
struct sg_entry {
	uint32_t ioaddr; /* This must actually be 32-bit! not caddr_t */
	uint32_t datalen; /* SPECIAL. 0x0000ffff is byte len.
			   *    0x80000000 is EOT signifier
			   */
};



/* This is for the PCI-IO space registers of the UDMA chipset */
/* This is memory mapped. do not change anything */
struct hpt_udmaregs {
	uint8_t command1;
	uint8_t filler1;
	uint8_t status1;
	uint8_t filler2;
	uint32_t scat_gath1;

	uint8_t command2;
	uint8_t filler3;
	uint8_t status2;
	uint8_t filler4;
	uint32_t scat_gath2;
};
/* Here are bit values used with the udma registers above*/
#define UDMACMD_DISKREAD	0x08
#define UDMACMD_START		0x01

#define UDMASTATUS_IRQ 0x04 /* write to clear*/
#define UDMASTATUS_ERR 0x02 /* write to clear */
#define UDMASTATUS_BUSY 0x01

/************************************************************/
/************************************************************/
/*            PCI "config register" values                  */
/************************************************************/
/************************************************************/

/* for timing0/timing1 register*/
#define	PCICONF_UDMAENABLE	(1<<28)
#define	PCICONF_UDMACYCLEBITS	(7<<16)
#define PCICONF_SHIFTBITS	(reg) ((reg& PCICONF_UDMACYCLEBITS)>>16)


/* defines for control2 */
#define PCICONF_FASTINTR_ENABLE (1<<15)
#define PCICONF_FLUSH_C0	(1<<10)
#define PCICONF_FLUSH_C1	(1<<11)

/****** defines for control3 and control4 registers ****/
/* "fast interrupt" setting for control3 and control4
 * 0 = interrupt when SG counter==512 bytes
 * 1 = interrupt when SG counter==1024 bytes
 * 2 = interrupt when SG counter==2048 bytes
 *
 * No, I dont understand this yet either, but note
 * the enable flag in reg control2
 * It should default to OFF.
 */
#define	PCICONF_FASTINTR	(3<<25)
#define	PCICONF_BYTESREMAINING	(15<<16) ?

/* 0x58: bus status1  (regular IDE status mapped through?)*/
#define	PCICONF_C1_DMARQ	(1<<5)
#define	PCICONF_C1_DMACK	(1<<4)
#define	PCICONF_C1_RDY		(1<<3)
#define	PCICONF_C0_DMARQ	(1<<2)
#define	PCICONF_C0_DMACK	(1<<1)
#define	PCICONF_C0_RDY		(1)


