/* Note: This source is divided into "UDMA", "FDISK", and "other"
 * sections. IDE stuff is now in hpt-ide.c
 */

/*
 * NASTY ASSUMPTIONS:
 *   1. we dont know what the hell to do with the TWO
 *      UDMA registers
 *   2. we're not sure we're doing the udma reg mapping properly any more
 *
 * This is VERY DIFFERENT from a usual disk driver, because normally,
 * controller and drive are in separate drivers. eg:
 * "scsi" vs "sd".
 * THIS driver however, takes care of both the bus-transfer handling,
 * and managing the two drives allowed on a single ide "bus"
 *
 * Maybe this is an ugly thing. But sun doesn't document nexus drivers,
 * so this is what I can do simply.
 *
 * version @(#) hpt.c 1.115@(#)
 */

#include <sys/errno.h>
#include <sys/conf.h>
#include <sys/file.h> /* for "FREAD" in ddi_copyout*/
#include <sys/modctl.h>
#include <sys/open.h>
#include <sys/stat.h>
#include <sys/kmem.h>
#include <sys/ksynch.h>

#include <sys/dkio.h>
#include <sys/vtoc.h>


#include <sys/dktp/fdisk.h>
#include <sys/dktp/dadkio.h>

#include <sys/ddidmareq.h>

/* THESE MUST BE LAST, of the sys includes */
#include <sys/ddi.h>
#include <sys/sunddi.h>

#include "defs.h"
#include "hpt-udma.h"
#include "hpt-queues.h"
#include "hpt-partitions.h"

/* just to make the compiler happy... 
 * ( reset_drive() is in this file, lower down)
 */
int reset_drive(struct hpt_softstate *,int drive );

/*Our globally visible stuffs, that we populate the global structs with */
static	int	hpt_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg,
		    void **result);
#ifdef PROBE
static	int	hpt_probe(dev_info_t *dip);
#endif
static	int	hpt_attach(dev_info_t *, ddi_attach_cmd_t);
static	int	hpt_open(dev_t *dev, int openflags, int otyp, cred_t *credp);

static	int	hpt_strategy(struct buf*);
static	int	hpt_devprint(dev_t,char *);

static	int	hpt_read(dev_t dev, struct uio* ustr, cred_t * credp);
static	int	hpt_write(dev_t dev, struct uio *uiop, cred_t *credp);

static	int	hpt_close(dev_t dev, int , int, cred_t *);



static	int	hpt_ioctl(dev_t dev, int cmd, intptr_t arg, int flags,
		    cred_t *credp, int *rvalp);

static	int	hpt_detach(dev_info_t *, ddi_detach_cmd_t );

static int hpt_aread(dev_t dev, struct aio_req *aio, cred_t *cred_p);
static int hpt_awrite(dev_t dev, struct aio_req *aio, cred_t *cred_p);

static	int	hpt_prop(dev_t, dev_info_t *, ddi_prop_op_t, int,
    char *, caddr_t, int *);

/* yes, this must be down here, because it uses some of
 * the function defs above
 */

#include "hpt-struct.h"

#include "hpt-ide.h"


int debug_level; /* multiple levels */
int waitforintr; /* enable slow "wait-for-interrupt" hack */
int merge_requests; /* optimize queue by merging adjacent I/O requests */
int udma_max; /* default of 10 gets set in hpt_attach.
		 Otherwise, may set max UDMA speed for all drives
		*/

int num_retries; /* times to retry a (probably bogus)DMA error */


/* This will print out stuff if debug_level set to 1 or higher */
static void routineprint(char *string){
	DLPRINT2(2,"hpt: In routine %s", string);
}

/* if you dont know the instance, this can be used to print a
 * debug message. Otherwise, use hpt_iprint()
 */
void hpt_print(char *message){
	DLPRINT2(1,"hpt: %s",message);
}
/*
 * This will only print out at debug_level 1 or higher 
 */
void hpt_iprint(int instance,char *message){
	DLPRINT3(1,"hpt%d: %s",instance,message);
}


#ifndef __GNUC__
#define USE_VPRINT /* gcc 2.81 is BROKEN, on my S7 x86! */
		/* some whining about __builtin_va_alist. sigh */
#endif

#ifdef USE_VPRINT
static void hpt_vprint(char *fmt,...){

	va_list ap;

	va_start(ap,fmt);
	vcmn_err(CE_NOTE,fmt,ap);
	va_end(ap);
}
#endif



/* cb_opts actually wants this?!! okay... */
static int hpt_devprint(dev_t dev,char *message){
	int instance=GETINSTANCE(getminor(dev));

	hpt_iprint(instance,message);
	return 0;
}

static kmutex_t   hpt_global_mutex;
void *hpt_globalstate; /* special pointer we need to keep around,
			* because ddi_get_soft_info() is wierd
			*/

static int intr_count, udmaintr_count;

/* This is called when you add_drv, and the module is loaded into
 * the kernel space.
 * However, hpt_attach is what is called to actually
 * add an instance of a device
 */
int
_init(void)
{
	int ret;

	routineprint("_init");
	/*
               * Initialize global mutex before mod_install'ing driver.
               * If mod_install() fails, must clean up mutex initialization
               */
	mutex_init(&hpt_global_mutex, NULL, MUTEX_DRIVER, (void *)NULL);

	ret=ddi_soft_state_init(&hpt_globalstate, sizeof(struct hpt_softstate),
			       HPT_NUMBUSES * HPT_NUMDRIVES);
	if(ret!=0){
		return (ret);
	}

	if ((ret = mod_install(&modlinkage)) != 0) {
		routineprint("init: modinstall !=0");

		ddi_soft_state_fini(&hpt_globalstate);
		mutex_destroy(&hpt_global_mutex);

		cmn_err(CE_NOTE, "(hpt) _init FAILED\n");
		return DDI_FAILURE;
	}

	intr_count=udmaintr_count=0;

	DLPRINT(1,"(hpt) _init returning SUCCESS, supposedly\n");
	return (DDI_SUCCESS);
}

int
_info(struct modinfo *modinfop)
{
	
	DLPRINT(1,"(hpt) _info called");
	return (mod_info(&modlinkage, modinfop));
}

/* This is called when you rem_drv  this driver. 
 * So this is for ALLL InSTANCES. Although hpt_detach() also gets
 * called.
 * It is the opposite of _init()
 */
int
_fini(void)
{
	int ret;

	DLPRINT(1, "(hpt) _fini called\n");
        /*
         * If mod_remove() is successful, we destroy our global mutex,
	 * etc, etc.
         */
	ret = mod_remove(&modlinkage);
	if ((ret != 0)){
		cmn_err(CE_NOTE, "hpt_fini: mod_remove failed");
		return (ret);
	}
	ddi_soft_state_fini(&hpt_globalstate);
	mutex_destroy(&hpt_global_mutex);

	return (ret);
}





/**********************************************************************/
/**********************************************************************/
/* ANd Here's the "everything else" Section! */
/**********************************************************************/
/**********************************************************************/

struct waitpkg {
	kmutex_t *mptr;
	kcondvar_t *cvptr;
};

/* This is used by the next function, intr_sleep().
 * Forget it exists.
 */
static void timeoutfunc(void *arg){
	struct waitpkg *ptr = (struct waitpkg*)arg;
	mutex_enter(ptr->mptr);
	cv_signal(ptr->cvptr);
	mutex_exit(ptr->mptr);
}

/* This function is meant to be a "interrupt-context-safe" method of
 * doing delay(), WITHOUT doing busywait
 * I dont understand why delay() isnt written this way.
 * Unfortunately, delay() isnt safe for interrupt context, according to
 * the manpage.
 *
 * Note that you probably dont want to call this with anything smaller than
 * 100 usecs, on a 400mhz cpu (Just a guess)
 * This routine itself probably has an overhead of 10 usecs, or something
 * ugly. Just another WAG, there.
 *
 *    ----- =====  WARNING ===== -----
 *  Do NOT use this, or timeout(), any time that you are using
 *  soft interrupts. There's some wierd interaction between timeout()
 *  and soft interrupts, it seems.
 *  (It may hang part of your system until the timeout triggers)
 */
void intr_sleep(int usecs){
	kmutex_t tmp_mutex;
	kcondvar_t tmp_wait;
	struct waitpkg pkg;

	pkg.mptr=&tmp_mutex;
	pkg.cvptr=&tmp_wait;

	mutex_init(pkg.mptr,NULL,MUTEX_DRIVER,NULL);
	cv_init(pkg.cvptr, NULL,CV_DRIVER,NULL);


	mutex_enter(pkg.mptr);
	timeout(timeoutfunc, &pkg, drv_usectohz(usecs));
	cv_wait(pkg.cvptr,pkg.mptr);
	mutex_exit(pkg.mptr);

	mutex_destroy(pkg.mptr);
	cv_destroy(pkg.cvptr);
}

/* NON-INTERRUPT_SAFE sleep
 * Use this in preference to intr_sleep, if you can
 * Odd thing is, it seems to be off by a factor of 10 on my Abit BP6 
 * So if you want to sleep 1 second, call sleep(100);
 * Oddly, changing to use drv_usectohz(100) does NOT do the same thing!
 */
void sleep(clock_t milisecs){
	clock_t onemilli=drv_usectohz(1000);
	delay(milisecs * onemilli);
}

/* called by anything that generates an interrupt, and expects to
 * wait for it on the cv
 * WE AQUIRE the wait_mutex here.
 * But NOT the instance_mutex, or bus mutex
 * return 0 if interrupt gotten, or 1 if timeout
 */
int waitforcv(struct hpt_softstate * softstate,int bus, int getmutex){
	clock_t timeout,now;
	timeout=drv_usectohz(8000000); /* 8 seconds */
	now=ddi_get_lbolt();

	if(getmutex==1)
		mutex_enter(&softstate->wait_mutex[bus]);

	if(cv_timedwait(&softstate->irq_wait[bus],
			&softstate->wait_mutex[bus],
			now+timeout) <1){

				mutex_exit(&softstate->wait_mutex[bus]);
			return 1;
	}
	mutex_exit(&softstate->wait_mutex[bus]);
	return 0;
}


/* This routine is called from hpt_detach to clean up all allocated stuff.
 * It is also called from hpt_attach, if an error is hit. So it has to
 * auto-detect which stuff has been allocated.
 * This is possible because softstate is zalloc'd.
 * Even though I could probably assume certain things, I try to make
 * NO assumptions here, to ensure the most robust cleanup
 * possible. We dont want to crash!
 */
static void cleanup(struct hpt_softstate *softstate){
	int instance, bus,drive;
	if(softstate==NULL){
		hpt_print("cleanup given NULL softstate? That's odd...");
		return;
	}
	instance=ddi_get_instance(softstate->dip);

	softstate->detach_start=1;
	mutex_enter(&softstate->instance_mutex);
		/* entering is just to wait until everyone else is done
		 * with them.
		 */
	for(bus=0; bus<HPT_NUMBUSES; bus++){
		DLPRINT(1,"cleanup - entering bus mutex, to destroy it");
		mutex_enter(&softstate->bus_mutex[bus]);
		mutex_exit(&softstate->bus_mutex[bus]);
		mutex_destroy(&softstate->bus_mutex[bus]);
		DLPRINT(1,"  destroyed bus mutex. doing others now");
		mutex_enter(&softstate->activequeue_mutex[bus]);
		mutex_exit(&softstate->activequeue_mutex[bus]);
		mutex_destroy(&softstate->activequeue_mutex[bus]);

		mutex_destroy(&softstate->wait_mutex[bus]);
		cv_destroy(&softstate->irq_wait[bus]);
		sema_destroy(&softstate->activequeue_sema[bus]);
		sema_destroy(&softstate->timeout_sema[bus]);

		if(softstate->hpt_ide_command_handle[bus] != NULL)
		  ddi_regs_map_free(&softstate->hpt_ide_command_handle[bus]);

		if(softstate->hpt_ide_control_handle[bus]!=NULL)
		  ddi_regs_map_free(&softstate->hpt_ide_control_handle[bus]);

	}

	for(drive=0;drive<HPT_NUMDRIVES;drive++){
		mutex_destroy(&softstate->waitqueue_mutex[drive]);
	}


	DLPRINT(1,"cleanup - starting to do pci teardown");
	if(softstate->pci_conf_h != NULL)
		pci_config_teardown(&softstate->pci_conf_h);


	if(softstate->hpt_udmaregs_handle != NULL)
		ddi_regs_map_free(&softstate->hpt_udmaregs_handle);

	mutex_exit(&softstate->instance_mutex);
	mutex_destroy(&softstate->instance_mutex);

	ddi_soft_state_free(hpt_globalstate, instance);
	DLPRINT(1,"hpt: cleanup DONE!");
}

/* because IDENTIFY string is in wierd order, we translate 
 * what we have stored from the drive in raw form
 * 'outstring' needs to be at least 21 chars
 * We also cut off the disk name after the first space, because
 * the IBM drive Im testing on gives wierd values after the first space
 */
void hpt_dname_to_string(uint16_t *array,char *outstring){
	char *parse;
	outstring[0]=(array[27]>>8)&0xff;
	outstring[1]=array[27]&0xff;
	outstring[2]=(array[28]>>8)&0xff;
	outstring[3]=array[28]&0xff;
	outstring[4]=(array[29]>>8)&0xff;
	outstring[5]=array[29]&0xff;
	outstring[6]=(array[30]>>8)&0xff;
	outstring[7]=array[30]&0xff;
	outstring[8]=(array[31]>>8)&0xff;
	outstring[9]=array[31]&0xff;
	outstring[10]=(array[32]>>8)&0xff;
	outstring[11]=array[32]&0xff;
	outstring[12]=(array[33]>>8)&0xff;
	outstring[13]=array[33]&0xff;
	outstring[14]=(array[34]>>8)&0xff;
	outstring[15]=array[34]&0xff;
	outstring[16]=(array[35]>>8)&0xff;
	outstring[17]=array[35]&0xff;
	outstring[18]=(array[36]>>8)&0xff;
	outstring[19]=array[36]&0xff;
	outstring[20]='\0';
	for(parse=outstring; *parse!='\0'; parse++){
		if(*parse==' '){
			*parse='\0';
			return;
		}
	}
}

/* This prints out  the drive identity string , when
 * you pass in the appropriate dev_info[] array
 */
static void print_identity(uint16_t *array)
{
	int count;
	char name[30];
	uint32_t numsect;

	hpt_dname_to_string(array,name);
	
	DLPRINT2(1,"hpt: identify: '%s'",	name);

	for(count=0;count<=4;count++){
		if(array[88] & (0x01 <<count)){
			DLPRINT2(1,"hpt:  UDMA %d supported",count);
		}
	}

	for(count=0;count<=4;count++){
		if(array[88] & (0x01 <<(count+8))){
			DLPRINT2(1,"hpt: UDMA %d enabled",count);
		}
	}

	numsect=(array[58]<<16) | array[57];
	cmn_err(CE_NOTE,"  number of sectors for device:  %d",
		numsect);

	/* This is the "official" capacity of the drive, as per Hale Landis*/
	/* Otherwise, [1]*[3]*[6] is capacity. aka C*H*S */
	numsect=(array[61]<<16) | array[60];
	cmn_err(CE_NOTE,"  number of USER sectors for device:  %d",
		numsect);


	cmn_err(CE_NOTE, "  number of logical cylinders: %d",array[1]);
	cmn_err(CE_NOTE, "  number of logical heads: %d",array[3]);
	cmn_err(CE_NOTE, "  number of logical sect/track: %d",array[6]);

	cmn_err(CE_NOTE, "  number of 'current' cylinders: %d",array[54]);
	cmn_err(CE_NOTE, "  number of 'current' heads: %d",array[55]);
	cmn_err(CE_NOTE, "  number of 'current' sect/track: %d",array[56]);


	count=array[75]&0x1f;
	cmn_err(CE_NOTE,"  queue depth-1 == %d",count);
	cmn_err(CE_NOTE,"  atapi flags == %x", array[80]&0xf);
	if(array[83] & 0x2) {
		cmn_err(CE_NOTE,"  83=QUEUED commands supported");
	} else {
		cmn_err(CE_NOTE,"  83=QUEUED commands NOT supported");
	}

	if(array[86] & 0x2) {
		cmn_err(CE_NOTE,"  86=QUEUED commands supported");
	} else {
		cmn_err(CE_NOTE,"  86=QUEUED commands NOT supported");
	}
	if(array[85] & 0x20) {
		cmn_err(CE_NOTE,"  write cache ON");
	}
	if(array[85] & 0x40) {
		cmn_err(CE_NOTE,"  look-ahead enabled");
	}

}


/* Used in attach, AND alloc_sg_table. Initialized in attach */
ddi_device_acc_attr_t dev_attr;




/****************************************************************
 * timeout_intr() 
 * This is a "soft interrupt" handler routine. It checks to see
 * if there was a timeout waiting for an I/O to process.
 *
 * It is initially invoked when a job is put on the queue, and
 * a timeout thread is not running already.
 * (judged by if the timeout_sema[bus] semaphore is held)
 *
 * All it does is sleep for a while, then see if the queue has changed.
 * It's a cut down version of waitforactivechanged.
 * Except this handles timeout for all disks in this instance.
 * Additionally, after scanning all disk queues once, if there is even
 * one "active" job.. it waits for a bit, and runs the loop again.
 *
 * In principle, I dont like sitting here, without returning
 * DDI_INTR_[CLAIMED/UNCLAIMED] as an interrupt handler.
 * It feels very wrong to me.
 *
 * Also... , there IS a small race condition here. In theory, we could detect
 * nothing to watch, and prepare to release semaphore.
 * Then, something is added, but a new timeout is not started, because
 * the semaphore is already held by a running timeout thread.
 * Meanwhile, the current one finishes releasing the semaphore.
 *
 * I DONT CARE, because the next access to the disk will kick off another
 * timeout thread, so it wont be a permenant hang.
 * I believe semaphores are better than adding a mutex or something here.
 */
uint_t timeout_intr(caddr_t voidptr){
	int instance=(int)voidptr;
	int bus=0,dcount,workneeded;
	struct hpt_softstate * softstate;
	uint_t returnval=DDI_INTR_UNCLAIMED;
	struct buf*bp[HPT_NUMDRIVES];

	softstate=ddi_get_soft_state(hpt_globalstate, instance);
	if(softstate==NULL){
		printf("ERROR: timeout_intr could not get softstate");
		return returnval;
	}

	/* Technically, this may be somewhat redundant, as it might
	 * not be possible to trigger a soft interrupt when one is already
	 * being processed. (they may serially fire).
	 * But KEEP THIS IN HERE anyway!!!
	 */
	sema_p(&softstate->timeout_sema[bus]);

	workneeded=0;
	for(dcount=0;dcount<HPT_NUMDRIVES;dcount++){
		bp[dcount]=softstate->waitqueue[dcount].b_forw;
		if(bp[dcount] !=NULL)
			workneeded=1;
	}
	if(workneeded==0){
		DLPRINT(2,"timeout_intr: called with no work needed");
		sema_v(&softstate->timeout_sema[bus]);
		return returnval;
	}

	/* Okay, work IS needed. So this trigger was for us */
	returnval=DDI_INTR_CLAIMED;

	/* Now sleep for a while, and see if things have cleared up.
	 * if not, then figure out which disk(s) need cleaning up, and trigger
	 * drive resets for em
	 */

	/* intr_sleep is in microseconds*/
	DLPRINT(1,"timeout_intr sleeping now");
	intr_sleep(20* 1000000);
	DLPRINT(1,"timeout_intr woke up");

	/* need to check for TWO things here:
	 * a) does something need resetting
	 *     [if so, do drive reset]
	 * b) do we need to loop and keep monitoring?
	 *     [if so, set workneeded==1]
	 */
	workneeded=0; dcount=0;
	while (1){
		/* first check to see if something is stuck*/
		if(bp[dcount]!=NULL){
			if(bp[dcount]==softstate->waitqueue[dcount].b_forw){
				printf("ERROR: resetting disk %d",dcount);
				if(software_reset(softstate,bus)!=0){
					printf("ERROR: could not reset bus!");
				}
				/* now re-enable both drives */
				init_udma(softstate,bus<<1);
				init_udma(softstate,(bus<<1)+1);
			}
		}

		/* Now, just see if there's a reason for us to keep looping.
		 * Save state for next time in the loop, and examine now.
		 */
		bp[dcount]=softstate->waitqueue[dcount].b_forw;
		if(bp[dcount]!=NULL) {
			workneeded=1;
		}
		
		dcount++;
		if(dcount>=HPT_NUMDRIVES){
			dcount=0;
			if(workneeded==0){
				break;
			}
			DLPRINT(1,"timeout_intr REsleeping now");
			intr_sleep(20* 1000000);
			DLPRINT(1,"timeout_intr woke up again");
		}
	}

	/* All done, and no more work pending.
	 * increment the timeout lockout semephore before exiting
	 */
	sema_v(&softstate->timeout_sema[bus]);


	return returnval;
}

/* Used as a jump point if we use timeout() as a watchdog for I/O requests.
 * Call with ( (instance<<8)|(drivenum&x0f) )
 *
 * If we're here, then an I/O took too long to complete.
 *
 * Note that this is like interrupt context.
 */
void hpt_timeout_handler(void *arg){
	int param,instance,drive,bus;
	struct hpt_softstate * softstate;

	DLPRINT(0,"hpt TIMEOUT handler called");

	param=(int)arg;
	drive=param&0xf;
	instance=(param&0xf0)>>8;

	bus=drive>>1;
	
	softstate=ddi_get_soft_state(hpt_globalstate, instance);
	if(softstate==NULL){
		printf("ERROR: timeout_handler couldnt get softstate??");
		return;
	}

	printf("DEBUG: timeout handler getting mutex, sleeping first");

	drv_usecwait(1000000);

	mutex_enter(&softstate->bus_mutex[bus]);

	printf("DEBUG: timeout handler GOT mutex, doing cleanup.");

	hpt_cleanup_active(softstate,drive,EIO);

/* Originally, tried to get the hardware back to full working order.
 * But now I'm thinking this is not the best place to try
 * to do that.
 * Maybe hpt_start_one_job would be better.
 * Right now, we're sort of in soft_interrupt context.
 */
#ifdef NOTNOW	
	/* remember, this will trigger interrupt, which will clean up things*/
	if(software_reset(softstate,bus)!=0){
		printf("ERROR: reset_drive failed from timeout handler??");
	}
	/* now re-initialize drive */
	init_udma(softstate,drive);
#endif

	printf("DEBUG: timeout handler releasing mutex");
	mutex_exit(&softstate->bus_mutex[bus]);
	DLPRINT(1,"Timeout handler done");

}


/* Send IDE NOP command. This will trigger an interrupt, if
 * interrupts are enabled. Which of course is a great way to check
 * if our interrupt handler is registered
 *
 * If feature register==0, clear queued requests.
 * If feature register==1, do NOT stop any queued requests.
 *
 * Either way, will set ERR flag on completion, and ABRT bit in error reg.
 * Up to caller to clear those bits.
 *  (will be seen by interrupt handler, if interrupt enabled!)
 *
 * I'm going to try using this as a "device reset" for regular devs
 * (since DEVICE_RESET is only for "PACKET" devices)
 *
 * If interrupts are enabled, set with_intr==1, else 0.
 */
void do_nop(struct hpt_softstate *softstate, int bus, int with_intr){
	/* This is for hard interrupt testing */
	/* But we want it in HERE, not hpt_attach, so it is
	 * easier to handle crash-reboots!
	 */

	if(with_intr==1){
		if(idecmd_checkstat(softstate,bus,IDECMD_NOP) !=0){
			DLPRINT(1,"do_nop: command NOT SENT?");
		} else {
			DLPRINT(1,"Done NOP IDEcmd okay. ");
		}
		return;
	}
	/* else */
	if(idecmd_nocheck(softstate,bus,IDECMD_NOP) != 0){
		DLPRINT(1,"do_nop: command NOT SENT?");
		return;
	}
	if(waitforbusy(softstate,bus) !=0){
		DLPRINT(1,"do_nop: ERROR: busy never seen");
		return;
	}
	if(wait_while_busy(softstate,bus)!=0){
		DLPRINT(1,"do_nop: BUSY got stuck!!");
		return;
	}
	DLPRINT(1,"do_nop expects err flag. Double-read:");
	printstatus(readstatus(softstate,bus));
	readaltstatus(softstate,bus);
	DLPRINT(1,"  second read:");
	printstatus(readstatus(softstate,bus));

}

/* hpt_handle_intr gets called by the system, when an interrupt
 * is generated. ANY interrupt the systems thinks we might be intersted in.
 *
 * "voidptr" is simply our instance number, as mentioned 
 * by ddi_add_intr()
 *
 *
 * If interrupt is not for either bus, return DDI_INTR_UNCLAIMED
 *
 * NOTE: This is in "Interrupt context", NOT USER-CONTEXT
 * NOTE THAT WE CANNOT USE cmn_err in here!!!
 *
 * MUTEXES:
 *   We  grab bus_mutex[bus] , call some stuff, and release just
 *   before exit
 */

uint_t
hpt_handle_intr(caddr_t voidptr){
	int instance=(int)voidptr;
	int bus=0, errstatus=0, ideerrflag=0,disk=0;
	struct hpt_softstate * softstate;
	struct sgclump *clumpptr;
	uint8_t tmpstatus;
	int activeerr;

	/* global counter */
	intr_count++;

	softstate=ddi_get_soft_state(hpt_globalstate, instance);
	if(softstate==NULL)
		return DDI_INTR_UNCLAIMED;

	DLPRINT(1,"hpt_handle_intr: starting by getting mutex");
	mutex_enter(&softstate->bus_mutex[bus]);

	/* do explicit check INSIDE mutex.
	 * used to be before mutex, but I think that could lead
	 * to double-trigger problems
	 */
	if(!got_udma_intr(softstate,bus)){
		DLPRINT(1,"hpt_handle_intr: UDMA status clear; spurious interrupt");
		mutex_exit(&softstate->bus_mutex[bus]);
		return DDI_INTR_UNCLAIMED;
	}


	if(softstate->timeout_id[bus] !=0){
		untimeout(softstate->timeout_id[bus]);
		DLPRINT2(2,"called untimeout on %d",
			 (int)softstate->timeout_id[bus]);
		softstate->timeout_id[bus]=0;
	}

	DLPRINT(2,"hpt_handle_intr: have bus mutex");
	if(debug_level>1){
		hpt_printactivequeue(softstate,disk);
	}

	stop_udma(softstate,bus);
	
	{
		/* This may screw things up since we havent turned off ATA yet*/
		int remaining;
		remaining=hpt_udma_getremaining(softstate,bus);
		DLPRINT2(1,"DEBUG: remaining bytes after intr: %d\n",remaining);
	}
	
	if(got_udma_err(softstate,bus)){
		int remaining=hpt_udma_getremaining(softstate,bus);
		/* dont bother normally to print to console, because half the
		 * time
		 * the OS will retry the write anyway, and it will
		 * successfully work!
		 */
		DLPRINT(1,"hpt_handle_intr: UDMA ERR DETECTED with interrupt");
		DLPRINT2(1,"   %d bytes remaining in udma buffer\n",
		       remaining);

		errstatus=EIO;
		clear_udma_err(softstate,bus);
	}

	/* This useless read of the IDE status register,
	 * enables clearing the IRQ flag in UDMA regs,
	 * and stops it retriggering.
	 * Useless, unless there's an error showing, that is.
	 */
	tmpstatus=readstatus(softstate,bus);

	if(tmpstatus & IDESTATUS_ERR){
	
		printf(" -- hpt_handle_intr detected IDE status err\n");
		printideerr(readerrval(softstate,bus));
		ideerrflag=1;
		errstatus=EIO;
	}

	if(got_udma_intr(softstate,bus)){
		/* This turns out to be the usual place the flag is
		 * "really" cleared.
		 */
		clear_udma_intr(softstate,bus);
	}

	/* NOW we can touch the IDE registers */
	if(wait_while_drq(softstate,bus)!=0){
		DLPRINT(1,"hpt ERROR: DRQ not clearing after interrupt");
	}

	/* First find out which drive triggered the interrupt*/
	disk=ide_which_drive(softstate,bus);

	/* This call takes the job off the active queue,
	 *  calls biodone() on the job,
	 * and increments the queue SEMAPHORE.
	 * It does nothing with the bus,
	 * so we dont need the bus mutex!
	 *
	 * ...Except that I've been having mysterious problems,
	 * so we're back in paranoia mode
	 */

	clumpptr=hpt_getactivesgclump(softstate,bus);

	/* This is atomic,I think*/
	udmaintr_count++;
	if(clumpptr==NULL){
		DLPRINT2(1,"hpt_handle_intr: drive %d trigged, with no active ptr",
			 disk);
			
		mutex_exit(&softstate->bus_mutex[bus]);
		return DDI_INTR_CLAIMED;
	}

	/* TEMPORARY DEBUG PRINT HERE */
	/*printsgtable(clumpptr);*/

	if(errstatus==0){
		/* Normal, "nice" cleanup */
		DLPRINT(1,"hpt_handle_intr calling cleanup_active");
		activeerr=hpt_cleanup_active(softstate,disk,errstatus);

		DLPRINT(1,"hpt_handle_intr releasing mutex in normal return");
		mutex_exit(&softstate->bus_mutex[bus]);

		/* MUST DO THIS AFTER MUTEX RELEASE, for cache reasons */
		/* This is the ONLY PLACE sema_v is called on activequeue */
		if(activeerr==0){
			DLPRINT(1,"hpt_handle_intr releasing activequeue SEMAPHORE");
			sema_v(&softstate->activequeue_sema[bus]);
		}
		
		return DDI_INTR_CLAIMED;
	}

	/* else */

	if(ideerrflag!=0){
		printf("   WHOA! IDE error!");
		activeerr=hpt_cleanup_active(softstate,disk,errstatus);
		mutex_exit(&softstate->bus_mutex[bus]);

		if(activeerr==0){
			DLPRINT(1,"hpt_handle_intr releasing activequeue SEMAPHORE");
			sema_v(&softstate->activequeue_sema[bus]);
		}

		return DDI_INTR_CLAIMED;
	}

	/* else */
	/* Retry the DMA operation, after tweaking
	 * For some inconceivable reason, DMA *writes*
	 * seem to fail 30% of the time.
	 * Legend has it the HPT366 has problems with "fast" drives.
	 *
	 */
	printf("   (DMA error)");

	/*
	printf("DEBUG: ERRed request is block %lld, len %ld",
			sg_clump->bufptr->b_lblkno,
			sg_clump->bufptr->b_bcount);
	*/

	DLPRINT(1,"DEBUG: Forcing retry_count==0");
	clumpptr->retry_count=0;

	if(clumpptr->retry_count==0){
		DLPRINT(0, "retry count=0. erroring out to cleanup_active");
		activeerr=hpt_cleanup_active(softstate,disk,errstatus);
		mutex_exit(&softstate->bus_mutex[bus]);
		if(activeerr==0){
			DLPRINT(0,"hpt_handle_intr releasing activequeue SEMAPHORE");
			sema_v(&softstate->activequeue_sema[bus]);
		}
		
		return DDI_INTR_CLAIMED;
	}

#ifdef DORETRY
		
	if(debug_level>0){
		DLPRINT(1,"hpt-handle-intr UDMA BEFORE retry starts:");
		printudmastatus(readudmastatus(softstate,bus));
		
	}
		
	if(realloc_sg_table(softstate, clumpptr)!=0)
	{
		printf("ERROR in realloc_sg_table\n");
		activeerr=hpt_cleanup_active(softstate,disk,errstatus);
		mutex_exit(&softstate->bus_mutex[bus]);
		if(activeerr==0){
			DLPRINT(0,"hpt_handle_intr releasing activequeue SEMAPHORE");
			sema_v(&softstate->activequeue_sema[bus]);
		}
		return DDI_INTR_CLAIMED;
	}

	/*
	DLPRINT(1,"hpt_handle_intr calling HACKERY!");
	hpt_hackery(softstate,disk);
	*/

	hpt_read_dma_polled(softstate,disk,(uint64_t)10000);

	DLPRINT(1,"  Now doing REAL RETRY!");
	enable_interrupt(softstate,bus);

	if(install_sgtable(softstate,disk)!=0) {
		printf("ERROR in install sgtable\n");
		activeerr=hpt_cleanup_active(softstate,disk,errstatus);
		mutex_exit(&softstate->bus_mutex[bus]);
		if(activeerr==0){
			DLPRINT(0,"hpt_handle_intr releasing activequeue SEMAPHORE");
			sema_v(&softstate->activequeue_sema[bus]);
		}
		return DDI_INTR_CLAIMED;
	} else {
		DLPRINT(1,"hpt_handle_intr: re-inserted new sgtable");
	}
		
	clumpptr->retry_count -=1;
		
	if(hpt_start_io(softstate,disk)!=0){
		DLPRINT(0,"hpt_handle_intr: retry failed, calling cleanup_active");
		activeerr=hpt_cleanup_active(softstate,disk,EIO);
	} else {
		DLPRINT(1,"hpt_start_io returned okay ");
	}
#endif /************ DORETRY */


	mutex_exit(&softstate->bus_mutex[bus]);
	if(activeerr==0){
		DLPRINT(0,"hpt_handle_intr releasing activequeue SEMAPHORE");
		sema_v(&softstate->activequeue_sema[bus]);
	}

	return DDI_INTR_CLAIMED;

} /* end of hpt_handle_intr() */


/* Little utility used by wait_for_interrupt().
 * Just do drv_usecwait()s until either the active queue for the drive
 * changes from given value, or we timeout.
 * OR, we detect a failure
 *
 * return 0 on change, 1 on timeout
 */
int waitforactivechange(struct hpt_softstate *softstate,int disk,
		       struct sgclump *sgptr){
	int bus=disk>>1;

	DLPRINT(2,"now in waitforactivechange() ...");

	drv_usecwait(100);

	if(hpt_getactivesgclump(softstate,disk) != sgptr)
		return 0;
	if(got_udma_err(softstate,bus)){
		DLPRINT(1,"  UDMA err detected");
		return 1;
	}

	drv_usecwait(100);
	if(hpt_getactivesgclump(softstate,disk) != sgptr)
		return 0;
	if(got_udma_err(softstate,bus)){
		DLPRINT(1,"  UDMA err detected");
		return 1;
	}

	drv_usecwait(200);
	DLPRINT(2," Hmm.. interrupt is a little late...");
	if(hpt_getactivesgclump(softstate,disk) != sgptr)
		return 0;
	if(got_udma_err(softstate,bus)){
		DLPRINT(1,"  UDMA err detected");
		return 1;
	}

	DLPRINT(1,"hpt: oh-oh... looks like an interrupt timeout");
	drv_usecwait(1 * 1000);
	if(hpt_getactivesgclump(softstate,disk) != sgptr)
		return 0;
	if(got_udma_err(softstate,bus)){
		DLPRINT(1,"  UDMA err detected");
		return 1;
	}

	/* in theory, 10000 == 10 millisec */
	drv_usecwait(10000 * 1000);
	if(hpt_getactivesgclump(softstate,disk) != sgptr)
		return 0;
	else
		return 1;
} /*end of waitforactivechange */



/* Silly little debug routine that makes sure interrupt gets triggered
 * for bufptr current on top of active queue.
 * We just check for any change to the active ptr for this disk,
 * and take any change as s sign that "yes, an interrupt happened".
 *
 * If one does NOT happen in an acceptible amount of time, and something
 * is stuck on the active queue... this does a fairly dramatic cleanup
 * operation!! (After which... well, this driver is probably screwed)
 *
 * Called by hpt_strategy, if enabled in conf file.
 *
 */
void wait_for_interrupt(struct hpt_softstate *softstate,int bus,int disk)
{
	struct sgclump*sgptr=hpt_getactivesgclump(softstate,disk);
	DLPRINT(1,"wait_for_interrupt: start");

	if(sgptr==NULL){
		DLPRINT(1,"wait_for_interrupt() had nothing on queue");
		return;
	}

	if(waitforactivechange(softstate,disk,sgptr)==0){
		DLPRINT(2," request has cleared queue normally");
		return;
	}

	/***********************************************************
	 *    In ERROR CONDITION (timeout) now                     *
	 * if we're here, the active queue for the disk queue has not changed in *
	 * over 10 seconds. We're pretty durn sure it's not going  *
	 * to change again. So we dont bother to refresh value     *
	 ***********************************************************/
	printf("ERROR: wait_for_interrupt: interrupt failed.");

	if(debug_level>=0 ){
		printudmastatus(readudmastatus(softstate,bus));
	}
	if(got_udma_busy(softstate,bus)==1){
		DLPRINT(1," waitforinterrupt stopping UDMA engine\n");
		stop_udma(softstate,bus);
	}
	if(debug_level>=0 ){
		/* Remember, CANNOT DO THIS SAFELY until UDMA STOPPED!!*/
		printstatus(readaltstatus(softstate,bus));
	}


	DLPRINT2(1,"  intr_count is %d", intr_count);
	DLPRINT2(1,"  udmaintr_count is %d", udmaintr_count);

	/* We now send an IDE drive reset command.
	 * This actually triggers an INTERRUPT, with error flags set.
	 * At which point, after we release the mutex, our regular
	 * interrupt handler will fire, see that there is an error,
	 * and clean up using our "normal" channels.
	 *
	 * The problem here is if somehow we get an interrupt
	 * return for something LATER ON. then we might
	 * corrupt things badly.
	 */
	printf(" w_f_i: trying to reset drive %d", disk);
	mutex_enter(&softstate->bus_mutex[bus]);

	if(software_reset(softstate,bus) !=0){
		printf(" w_f_i: DANGER DANGER DANGER: disk would not reset");
		/* We are now in very nasty territory
		 * Who has the lock on the disk queue?
		 */
	}
	if(debug_level>0){
		printf("STATUS after software_reset:");
		printudmastatus(readudmastatus(softstate,bus));
		printstatus(readaltstatus(softstate,bus));
	}

	/* change drive to "primary", so "drive+1"==slave */
	disk&=0xe;
	
	/* now re-enable both drives */
	if(softstate->drive_speed[disk]!=-1)
		init_udma(softstate,disk);
	if(softstate->drive_speed[disk+1]!=-1)
		init_udma(softstate,disk+1);

	printf("  w_f_i: reset bus successfully. Releasing mutex");
	mutex_exit(&softstate->bus_mutex[bus]);
	/* MUST RELEASE HERE to let interrupt routine clean up */

	printf("  w_f_i: seeing if interrupt cleans up active");
	if(waitforactivechange(softstate,disk,sgptr)==0){
		printf("  w_f_i: yeup, active queue cleaned");
	} else {
		printf("  w_f_i: detects NO CLEANUP of active queue. Doing now.");
		hpt_cleanup_active(softstate,disk,EIO);
		printf("  w_f_i: RECOMMEND FULL SYSTEM RESET");
	}

	if(debug_level>0){
		mutex_enter(&softstate->bus_mutex[bus]);

		printf(" w_f_i: printing register statuses");
		printstatus(readaltstatus(softstate,bus));
		printudmastatus(readudmastatus(softstate,bus));
		printf("  w_f_i: releasing bus mutex.");

		mutex_exit(&softstate->bus_mutex[bus]);
	}


	return;

}


/*
 * Strategy: ALL requests for I/O go through this routine
 * eventually. (even directly from external kernel routines)
 * We add the i/o request to the request queue, with  disksort()
 *
 * If the request queue was empty, we start processing it now.
 * kick off a fake interrupt, or call workloop by hand.
 *
 * Otherwise, we just assume that when the current request
 * gets done, we will automatically start the next one at that
 * time.
 *
 * Note that for some odd reason, strategy routines always return 0.
 * errors are flagged by calling bioerror()
 * On error, we always call biodone() as well.
 *
 */
int hpt_strategy(struct buf* bptr){
	struct hpt_softstate *softstate;
	int instance,minornum, disk,bus;
	int slice;
	int numblocks, slicenumblocks,startblock;

	DLPRINT(2,"entered hpt_strategy");

#ifdef FAKE_IT_ONLY
	if(bptr->b_flags & B_READ){
		printf("DEBUG: hpt_strat FAKING WRITE op\n");
		biodone(bptr);
		return 0;
	}
#endif
	minornum=getminor(bptr->b_edev);
	instance=GETINSTANCE(minornum);
	slice=minornum&DEV_SLICEMASK;
	softstate=ddi_get_soft_state(hpt_globalstate, instance);
	if(softstate == NULL){
		bptr->b_resid = bptr->b_bcount;
		hpt_iprint(instance,"strategy cannot get softstate");
		bioerror(bptr, EAGAIN);
		biodone(bptr);
		return (0);
	}
	disk=(minornum & DEV_DRIVEMASK) / DEV_DRIVEMULT;
	if(disk>1) bus=1; else bus=0;

	DLPRINT(2,"DEBUG: at of hpt_strategy,");
	DLPRINT2(2," intr_count is %d", intr_count);
	DLPRINT2(2," udmaintr_count is %d", udmaintr_count);

	bptr->b_private=NULL; /* CRITICAL for checks in other routines!*/
#ifdef DO_NOP
	/* check if interrupt handler is working */
	DLPRINT2(1,"  enabling interrupt for NOP on bus %d",bus);
	enable_interrupt(softstate,bus);
	do_nop(softstate,bus,1);
	DLPRINT(1,"  sleeping for a sec...");
	sleep(100);

	/*  -- NOTE -- : it is EXPECTED to get ERR/"ABRT" flag, as
	 * as per the ATA spec. So this "NOP" actually does something
	 */
	DLPRINT(1,"IDE status is:");
	printstatus(readstatus(softstate,bus));
	printudmastatus(readudmastatus(softstate,bus));
	DLPRINT2(1," intr_count is %d", intr_count);
	DLPRINT2(1," udmaintr_count is %d", udmaintr_count);

	bioerror(bptr, ENXIO);
	biodone(bptr);
	return (0);
#endif


	if(debug_level>0){
		uint64_t blkno=bptr->b_lblkno;
		printf(" hpt_strategy: slice %d, rel_startblock=0x%llx, bytecount=%ld, readflag=%d",
			slice,	blkno,
			bptr->b_bcount, bptr->b_flags& B_READ);
	}

	/* 16 minors per dev, 4 devs per instance. */
	if(debug_level>2){
		cmn_err(CE_NOTE,
			"hpt: strategy: minornum=%x, instance %d, disk %d",
			minornum, instance, disk);

		cmn_err(CE_NOTE,
		  "hpt: strategy has request for block %x, bytecount %d, bufsize %d",
		  bptr->b_lblkno, bptr->b_bcount, bptr->b_bufsize);
	}


	/* Get number of blocks in this I/O request. Truncate
	 * if requesting non-block-boundary amount of data.
	 */
	numblocks=(bptr->b_bcount)>>9;
	if(numblocks==0){
		hpt_print("strategy: hit wierd solaris 0-blockrequest. Setting to bufsize");
		numblocks=(bptr->b_bufsize)>>9;
		if(numblocks==0){
			hpt_print("STILL zero? ERROR");
		}
		
	}

	/* Now calculate the maximum number of blocks in the current slice,
	 * and also find the absolute disk block number for the start of
	 * the slice.
	 */

	if(minornum&HPT_FULLDISK){
		int fpart=minornum&0x7;

		startblock=getfstartblk(softstate,disk,fpart);

		slicenumblocks=getfblkcount(softstate,disk,fpart);


		DLPRINT3(2," for fpart %d, start of partition is %d",fpart,startblock);
		DLPRINT2(2," %d blocks available total",slicenumblocks);
		DLPRINT2(2," ... for relative block %lld",bptr->b_lblkno);
	} else {
		int sp_startblock; /* solaris fdisk partition startblock */
		sp_startblock=softstate->sol_startblk[disk];

		slicenumblocks=getsblkcount(softstate,disk,slice);

		DLPRINT2(2,"strategy: sp_startblock is supposedly %d",sp_startblock);

		/* Now apply relative positioning of solaris slice */
		 
		startblock=sp_startblock+
			softstate->partitions[disk].v_part[slice].p_start;

		DLPRINT3(2,"strategy:  relative blkno is %lld, count is %d",
			bptr->b_lblkno, numblocks);

	}

	/* and now calculate the block for the actual I/O request!! */
	startblock+=bptr->b_lblkno;

	DLPRINT3(2,"  hpt_strat: absolute start block=%d, len %d blocks",startblock,
			numblocks);

	if((bptr->b_lblkno <0 ) ||
	   (bptr->b_lblkno+numblocks > slicenumblocks) ||
	   (numblocks==0)   ){

		cmn_err(CE_WARN,"hpt: strategy failed: block %d past end of slice",
			bptr->b_lblkno);
		cmn_err(CE_WARN,"    slice %d has %d blocks",slice,slicenumblocks);


		bptr->b_resid=bptr->b_bcount;
		bioerror(bptr,EINVAL);
		biodone(bptr);
		return(0);
	}

	/*
	 * we reassign b_resid for disksort(), which allegedly sorts based
	 * on b_resid. (why sort on that?!!)
	 *
	 */
	bptr->b_lblkno = startblock;
	bptr->b_resid = bptr->b_lblkno;/* sorting field */

	
	bptr->av_forw=NULL;
	bptr->av_back=NULL;

	hpt_add_to_waitqueue(softstate,disk,bptr);




	/* This "test" actually activates physical I/O , puts request(s)
	 * on the active queue, and all that good stuff.
	 * I should rename it soon.
	 */
	if(hpt_start_one_job(softstate,disk) != 0){
		
		DLPRINT(0,"ERROR: hpt_strategy: hpt_start_one_job failed?!!");
		/* Note: hpt_start_one_job should have called bioerror
		 * and biodone already
		 */
		return 0; /* yes, return 0 even though ERROR happened */
	} /* else */

	if(waitforintr>0){
 		wait_for_interrupt(softstate,bus,disk);
	} else {

#ifdef SOFTINTR_TIMEOUT
		/* kick off a timeout soft_interrupt thread, IFF one isnt
		 * already running
		 */
		if(sema_tryp(&softstate->timeout_sema[bus])!=0){
			ddi_trigger_softintr(softstate->softintr_id);
			sema_v(&softstate->timeout_sema[bus]);
			DLPRINT(1,"started softintr timeout_intr thread");
		}
#endif
	}

	DLPRINT(2, "hpt: END of strategy() [after workloop]");

	/* We HAVE to return 0. Why this routine isn't type void, I dont know*/
	return (0);

}

/*
 * Write the vtoc associated with disk, to the solaris partition
 * on that disk.
 * We cheat, and just look at sol_startblk[disk]. If 0, fail.
 * Otherwise, write to sol_startblk[disk]+1
 * Yes, that's right, vtocs on Solaris Intel start at block 1.
 * This is just in case we are using the disk "raw", in which case
 * the fdisk label goes at block 0.
 * 
 * Note that "read_vtoc" is logically encapsulated in copy_fdisk_info
 *
 * Return 0 on okay, errno on fail (or 1==generic fail)
 * This belongs in hpt-partitons,but it wont load-link outside this file
 */
int hpt_write_vtoc(struct hpt_softstate*softstate,dev_t dev)
{
	int errnum=0;
	int minornum=getminor(dev);
	int bus=0;
	int disk=(minornum & DEV_DRIVEMASK) / DEV_DRIVEMULT;
	if(disk>1) bus=1;

#ifdef USE_DMA_FOR_IOCTL
	struct buf *mybuf;


	mybuf=getrbuf(KM_SLEEP);

	mybuf->b_flags=B_WRITE;
	mybuf->b_un.b_addr=(caddr_t)&softstate->partitions[disk];
	mybuf->b_lblkno=1;
	mybuf->b_bcount=512;
	mybuf->b_edev=dev;

	hpt_strategy(mybuf);
	biowait(mybuf);
	
	errnum=geterror(mybuf);
	freerbuf(mybuf);
#else  /* !USE_DMA_FOR_IOCTL */

	disable_interrupt(softstate,bus);
	if(hpt_write_pio(softstate,disk,
		     &softstate->partitions[disk],
		     softstate->sol_startblk[disk]+1, 1) != 0){
		cmn_err(CE_WARN,"write of VTOC info failed");
		errnum=EIO;
	}
	enable_interrupt(softstate,bus);
#endif   /* USE_DMA_FOR_IOCTL */

	
	return errnum;
}

/* stolen from bst.c */
static int hpt_aread(dev_t dev, struct aio_req *aio, cred_t *cred_p){
	return (aphysio(hpt_strategy, anocancel, dev, B_READ,
			minphys,aio));
}
static int hpt_awrite(dev_t dev, struct aio_req *aio, cred_t *cred_p){
	return (aphysio(hpt_strategy, anocancel, dev, B_WRITE,
			minphys,aio));
}


/* This routine pretty much exists solely to dynamically return
 * the 'nblocks' property in one form or another.
 */
static int
hpt_prop(dev_t dev, dev_info_t *dip,
	ddi_prop_op_t prop_op,int mod_flags,
	char *name,caddr_t valuep,int *lengthp)
{
	int nblocks;
	caddr_t		buffer=NULL;
	struct hpt_softstate * softstate;
	int instance=ddi_get_instance(dip);
	int minornum=getminor(dev);
	int slice=minornum&DEV_SLICEMASK;
	int drive=(minornum & DEV_DRIVEMASK) / DEV_DRIVEMULT;
	int returnsize=4;
	long long bytecount=0;

	softstate=ddi_get_soft_state(hpt_globalstate, instance);
	
	DLPRINT2(2,"hpt_prop called with request for %s",name);
	DLPRINT3(3,"  dev is of type %ld (0x%lx)",dev,dev);

	if (dev == DDI_DEV_T_ANY){
		DLPRINT(2," ignoring request for DDI_DEV_T_ANY");
		DLPRINT(2," passing prop call through to ddi_prop_op");
		DLPRINT2(2," hpt: passed property name is %s",	name);

		/* system default */
		return (ddi_prop_op(dev, dip, prop_op, mod_flags,
				    name, valuep, lengthp));
	}
	if((strcmp(name,"nblocks") !=0) &&
	   (strcmp(name,"Nblocks") !=0) &&
	   (strcmp(name,"size") !=0) &&
	   (strcmp(name,"Size") !=0))
	   {
		DLPRINT(1,"Unrecognized property request. Passing through");
		return (ddi_prop_op(dev, dip, prop_op, mod_flags,
				    name, valuep, lengthp));
	}
	   
	
	if(softstate==NULL){
		cmn_err(CE_WARN,"hpt_prop: couldnt get softstate??");
		return(DDI_PROP_NOT_FOUND);
	}
	if(drive>3){
		hpt_iprint(instance,"hpt_prop: drive num too large");
		return (DDI_PROP_NO_MEMORY);
	}

	if(prop_op==PROP_EXISTS){
		DLPRINT(2,"PROP_EXISTS check. Returningd true");
		return(DDI_PROP_SUCCESS);
	}
	/* FINALLY, we can actually DO something now!! */

	if(lengthp!=NULL){
	   DLPRINT3(3,"request for size of minor %x with bufferlength %d",
		    minornum,
		   *lengthp)
	}


	if(minornum&HPT_FULLDISK){
		nblocks=getfblkcount(softstate,drive,slice);
	} else {
		nblocks=getsblkcount(softstate,drive,slice);
	}

	if(strcmp(name, "size")==0){
		/* I THINK this should be just nblocks, but
		 * I need to confirm
		 */
		nblocks=(nblocks<<DEV_BSHIFT);
		returnsize=4;
	}else if(strcmp(name, "Size")==0){
		DLPRINT(3,"DOING FULL SHIFT LOAD of bytecount");
		bytecount=(nblocks<<DEV_BSHIFT);
		returnsize=8;
	} else if (strcmp(name, "nblocks")==0){
		returnsize=4;
	} else if (strcmp(name, "Nblocks")==0){
		bytecount=nblocks;
		returnsize=8;
	}

	DLPRINT2(3,"  (nblocks=%d)",nblocks);

	if(prop_op==PROP_LEN){
		/* They only want size, presumably so they can
		 * alloc something appropriate, and call again.
		 */
		DLPRINT2(2," Telling caller they need size=%d",returnsize);
		*lengthp=returnsize;
		return (DDI_PROP_SUCCESS);
	}

	if(prop_op== PROP_LEN_AND_VAL_ALLOC){
		DLPRINT(2," ALLOC'ing memory for value");
		/*
		 * Allocate buffer for caller
		 */
		buffer = kmem_alloc((size_t)returnsize,
			((mod_flags & DDI_PROP_CANSLEEP) ?
			KM_SLEEP : KM_NOSLEEP));
		if (buffer == NULL)  {
			cmn_err(CE_WARN,
			    "  no memory for alloc: returning fail");
			return (DDI_PROP_NO_MEMORY);
		}
		*(caddr_t *)valuep = buffer; /* Set callers buf ptr */
		*lengthp=returnsize;
	} else {
		/* else */
		if(valuep==NULL){
			DLPRINT(1,"ERROR: caller didn't give mem");
			return (DDI_PROP_NO_MEMORY);
		}
		buffer=valuep;
	}

	if(returnsize > (*lengthp)){
		DLPRINT(1,"caller buff too small");
		return (DDI_PROP_BUF_TOO_SMALL);
	}

	if(returnsize==8){
		DLPRINT(3,"  Copying FULL 8-byte Size val");
		bcopy(buffer, &bytecount, returnsize);
	} else {
		DLPRINT(3,"  Copying smaller nblocks val");
		bcopy(buffer, &nblocks, sizeof(nblocks));
	}

	DLPRINT(3,"Returning hpt_prop success");
	return (DDI_PROP_SUCCESS);

}


static int	hpt_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd,
			     void *arg, void **result){
	struct hpt_softstate *softstate;
	int instance;

	DLPRINT(2,"hpt_getinfo() called");

	/* Note: the ddi is broken, in some respects. According to manpage,
	 * we need to IGNORE the passed in *dip pointer.
	 */
	*result=NULL;
	instance=GETINSTANCE(getminor((dev_t) arg));

	switch(infocmd){
	    case DDI_INFO_DEVT2DEVINFO:
		softstate = (struct hpt_softstate *)ddi_get_soft_state(
					hpt_globalstate,instance);
		if (softstate == (struct hpt_softstate *)NULL) {
			cmn_err(CE_WARN, "hpt%d: getinfo: Bad soft state",
				instance);
			return (DDI_FAILURE);
		}
		*result=softstate->dip;
		return  DDI_SUCCESS;
		break;
	    case DDI_INFO_DEVT2INSTANCE:
		*result=(void*)instance;
		return DDI_SUCCESS;
		break;
	}
	return DDI_FAILURE;
}


#ifdef PROBE
static	int	hpt_probe(dev_info_t *dip){
	int target=-1, lun=-1;
   	routineprint("probe faked");
  	target=ddi_prop_get_int(DDI_DEV_T_ANY,dip,DDI_PROP_DONTPASS,
				"target", -1);
  	lun=ddi_prop_get_int(DDI_DEV_T_ANY,dip,DDI_PROP_DONTPASS,
				"lun", -1);
   
   	cmn_err(CE_NOTE, "hpt: probe- called fortarget=%d, lun=%d\n", target,lun);

	/* Normally, we would check the dev_info info,
	 * to see which device ID is being probed, and then
	 * check to see if the durn thing is actually there.
	 * But for now, assume .conf file is correct
	 */
   
	/* our two options for success are:
	 * DDI_PROBE_DONTCARE
	 * DDI_PROBE_SUCCESS
	 */
	return DDI_PROBE_DONTCARE;
	/*return DDI_PROBE_FAILURE;*/
}

#endif /* PROBE */

/*  We dont use a mutex here !!!
 * If you call it from anywhere other than hpt_attach, grab the
 * bus mutex first!
 *
 * Note that if you have disable_intr==1, it will first disable interrupts,
 * then ENABLE interrupts
 *
 * return 0 on okay, 1 on fail
 */
static int load_identify_info(struct hpt_softstate *softstate,int drivenum,
			      int disable_intr){
	uint8_t command;
	int retval=0;
	int bus=0;
	if(drivenum>1) bus=1;

/*	routineprint("load_identify_info");*/

	if(wait_while_busy(softstate,bus) !=0){
		DLPRINT(1,"hpt: load_identify busied out. Cannot even start");
		return 1;
	}

	if(disable_intr){
		disable_interrupt(softstate,bus);
	}
	if(ide_select_drive(softstate,drivenum) !=0){
		printf("load_identify_info could not select drive");
		return 1;
	}

	command=IDECMD_IDENTIFY;
	/* nocheckstat, because currently, getpiodatablock reads status*/
	
	if(idecmd_nocheck(softstate,0,command) !=0){
		DLPRINT2(1,"hpt: write of CMD_IDENTIFY on drive %d failed",
			 drivenum);
		return 1;
	}

	if(getpiodatablock(softstate,bus,softstate->dev_info[drivenum]) != 0){
		cmn_err(CE_NOTE, "getpidatablock returned error");
		retval=1;
	}

	DLPRINT(1,"   WARNING: hardcoding blkspersect");
	/* This hardcode assumes 1 sector==2048 bytes */
	softstate->blkspersect[drivenum]=4;
	/* number of sectors on drive * blkspersect == capacity in blocks */	

	hpt_print("getpiodatablock returned.");

	if(disable_intr){
		enable_interrupt(softstate,bus);
	}
	return retval;
}




/* This is called to enable/probe for a specific "instance" of
 * a device. In our case, one particular IDE BUS.
 * we have to make /device nodes for all physical devices on that
 * bus.
 */
static int hpt_attach(dev_info_t *dip, ddi_attach_cmd_t cmd){
   	int instance;
	int status, stat2,stat3;
	int primarygood=0,slavegood=0;
	int count,buscount; /* tmp counter */
	struct hpt_softstate *softstate;
	uint16_t reg16;
	uint32_t reg32;
	int bus=0;
	int nint;

		/* this is a HARDCODE right now, to make fixing later easier*/

   	instance=ddi_get_instance(dip);

	hpt_iprint(instance,"starting hpt_attach");

   	switch(cmd){
	 case DDI_RESUME:
	   	cmn_err(CE_NOTE, "hpt: Ignoring attach_RESUME\n");
		return DDI_FAILURE;
	 case DDI_PM_RESUME:
		cmn_err(CE_NOTE, "hpt: Ignoring attach_PM_RESUME\n");
		return DDI_FAILURE;
	 case DDI_ATTACH:
		dev_attr.devacc_attr_version = DDI_DEVICE_ATTR_V0;
		dev_attr.devacc_attr_dataorder = DDI_STRICTORDER_ACC;
		dev_attr.devacc_attr_endian_flags = DDI_STRUCTURE_LE_ACC;
	   	break;
	 default:
	   	cmn_err(CE_NOTE, "hpt: unknown attach command\n");
	   	return DDI_FAILURE;
	}
	if(instance!=0){
		hpt_iprint(instance,"DEBUG-ERROR: we only handle instance #0");
		hpt_iprint(instance," will not attach this instance");
		return (DDI_FAILURE);
	}


	/* This is kinda bad, since it will get reloaded if we have
	 * multiple instances.. oh well :-)
	 */
	debug_level=ddi_prop_get_int(DDI_DEV_T_ANY,dip, DDI_PROP_DONTPASS,
				    "debug-level", 0);
	waitforintr=ddi_prop_get_int(DDI_DEV_T_ANY,dip, DDI_PROP_DONTPASS,
				    "wait-for-interrupt", 0);
	merge_requests=ddi_prop_get_int(DDI_DEV_T_ANY,dip, DDI_PROP_DONTPASS,
				    "merge-requests", 0);

	num_retries=ddi_prop_get_int(DDI_DEV_T_ANY,dip, DDI_PROP_DONTPASS,
				    "num-retries", 5);
	udma_max=ddi_prop_get_int(DDI_DEV_T_ANY,dip, DDI_PROP_DONTPASS,
				    "udma-max", 10);

	if(debug_level>0){
		cmn_err(CE_NOTE, "hpt: debug-level is %d\n",debug_level);
	}

   	hpt_iprint(instance,"hpt: attach called");
	if(ddi_dev_nintrs(dip, &nint)!=DDI_SUCCESS){
		cmn_err(CE_NOTE,"hpt: ddi_dev_nintrs FAILED?");
	} else {
		DLPRINT2(1,"hpt; by the way, this device has %d interrupts",
			nint);
	}

	if (ddi_intr_hilevel(dip, 0) != 0) {
		cmn_err(CE_WARN,
		    "hpt %d: high-level interrupt not supported",
		    instance);
		return (DDI_FAILURE);
	}

	if (ddi_soft_state_zalloc(hpt_globalstate, instance) != DDI_SUCCESS) {
		cmn_err(CE_WARN, "hpt%d: Failed to alloc soft state",
		    instance);
		return (DDI_FAILURE);
	}
	softstate = (struct hpt_softstate *)ddi_get_soft_state(hpt_globalstate,
							   instance);
	if (softstate == (struct hpt_softstate *)NULL) {
		cmn_err(CE_WARN, "hpt%d: Bad soft state", instance);
		ddi_soft_state_free(hpt_globalstate, instance);
		return (DDI_FAILURE);
	}
	DLPRINT2(1,"hpt_attach: softstate is at %p",softstate);

	softstate->dip = dip;/* just for hpt_getinfo() */


	hpt_iprint(instance,"starting mutex init");

	if(ddi_get_iblock_cookie(dip, 0, &softstate->interrupt_cookie)
	   != DDI_SUCCESS){
		cmn_err(CE_WARN,
			"hpt%d: failed to get iblock_cookie",
			instance);
		cleanup(softstate);

		return (DDI_FAILURE);
	}



	/* It is not clear WHY I should pass the intr_cookie in here.
	 * But this is what the docs recommend, so...
	 */
	mutex_init(&softstate->instance_mutex, NULL,MUTEX_DRIVER,
		   NULL);

	for(buscount=0;buscount <HPT_NUMBUSES;buscount++){
		mutex_init(&softstate->bus_mutex[buscount], NULL,MUTEX_DRIVER,
			   (void*)softstate->interrupt_cookie);

		mutex_init(&softstate->activequeue_mutex[buscount], NULL,
			   MUTEX_DRIVER,  NULL);
		mutex_init(&softstate->wait_mutex[buscount], NULL,MUTEX_DRIVER,
			   NULL);

		cv_init(&softstate->irq_wait[buscount], NULL,CV_DRIVER,NULL);
		sema_init(&softstate->activequeue_sema[buscount], 1,NULL,
			  SEMA_DRIVER,NULL);
		sema_init(&softstate->timeout_sema[buscount], 1,NULL,
			  SEMA_DRIVER,NULL);

	}

	for(count=0;count<HPT_NUMDRIVES;count++){
		mutex_init(&softstate->waitqueue_mutex[count], NULL,
			   MUTEX_DRIVER, NULL);
	}

	if(pci_config_setup(dip, &softstate->pci_conf_h) != DDI_SUCCESS){
		hpt_iprint(instance, "could not get pciconf");
		ddi_soft_state_free(hpt_globalstate, instance);
		return (DDI_FAILURE);
	}

	reg32=pci_config_get32(softstate->pci_conf_h, 0);
	DLPRINT2(1, "hpt: PCI ID of device is %x", reg32);

	reg16=pci_config_get32(softstate->pci_conf_h, 4);
	if( (reg16&0x4) ==0){
		cmn_err(CE_WARN, "hpt: instance %d non-bus-master: fail",
			instance);
		cleanup(softstate);
		return DDI_FAILURE;
	}

#ifdef WHEREISIOSPACE
	reg32=pci_config_get32(softstate->pci_conf_h, 0x20);
	DLPRINT2(1, "hpt: IOspaceptr@pci20 is %x", reg32);
#endif

#ifdef WHEREISROM
	reg32=pci_config_get32(softstate->pci_conf_h, 0x30);
	DLPRINT2(1, "hpt: ROMspaceptr@pci30 is %x", reg32);
#endif
#ifdef WHATINTR
	reg16=pci_config_get8(softstate->pci_conf_h, 0x3C);
	DLPRINT2(1, "hpt: Interrupt [pci@3C] is %x", reg16);
	reg16=pci_config_get8(softstate->pci_conf_h, 0x3D);
	DLPRINT2(1, "hpt: INTA/B is %x", reg16);
#endif


	/* bus 0 */
	status=ddi_regs_map_setup(dip, 1,
		(caddr_t *)&softstate->ide_command_regs[0],
		0, 8,&dev_attr, &softstate->hpt_ide_command_handle[0]);

	stat2=ddi_regs_map_setup(dip, 2,
		(caddr_t *)&softstate->ide_control_reg[0],
		0, 4,&dev_attr, &softstate->hpt_ide_control_handle[0]);

	/* REMEMBER: unlike BX chipset: one instance, ONE bus.
	 * PLUS, control stuff above, is 4bytes, instead of 1!
	 */


	/* shared UDMA register set */
	stat3=ddi_regs_map_setup(dip, 3, (caddr_t *)&softstate->hpt_udmaregs,
		0, 0x10,&dev_attr, &softstate->hpt_udmaregs_handle);

	if((status != DDI_SUCCESS) || (stat2!= DDI_SUCCESS)
	   || (stat3 != DDI_SUCCESS)){
		cmn_err(CE_WARN,"could not map I/O regs");
		cleanup(softstate);
		return (DDI_FAILURE);
	}


	/* believe it or not, if we dont check for this first..
	 * we could potentially lock up the entire (intel) machine!!!
	 */
	if(got_udma_busy(softstate,bus)){
		stop_udma(softstate,bus);
	}

	/* Start with a clean slate, hardware wise */
	DLPRINT(1,"Doing 'software reset'");
	if(software_reset(softstate,0)!=0){
		hpt_iprint(instance,"couldnt do software reset");
		return DDI_FAILURE;
	} else {
		hpt_iprint(instance,"software reset happened okay");
	}

	softstate->queued_ok[0]=1; /* init_udma unsets this if not good */


	/* Now go probe each potential device for capabilities, and to see if
	 * there is even something THERE
	 */
	DLPRINT(2,"testing primary device");

	if(load_identify_info(softstate, 0,0)!=0){
		DLPRINT(1, "hpt: load_identify failed primary");
		softstate->drive_speed[0]=-1;
	} else {
		primarygood=1;
		DLPRINT(1, "hpt: Completed identify load on primary");
		if(debug_level>1){
			print_identity(softstate->dev_info[0]);
		}

		DLPRINT(1, "hpt: trying for disk partition on primary");
		/* copies fdisk AND vtoc info into memory */
		if(get_fdiskpart(softstate,0)!=0){
			hpt_iprint(instance, "No Fdisk info for primary");
		}
		/* get_fdiskpart() leaves interrupts enabled on exit*/
		disable_interrupt(softstate,bus);

	}

	DLPRINT(1,"testing slave device");
	if(load_identify_info(softstate, 1,0)!=0){
		hpt_iprint(instance,"load_identify_info on slave failed");
		softstate->drive_speed[1]=-1;
	} else {
		slavegood=1;
		DLPRINT(1, "hpt: Completed identify load on slave");
		if(get_fdiskpart(softstate,1)!=0){
			hpt_iprint(instance, "No Fdisk info for slave");
		}
		/* get_fdiskpart() leaves interrupts enabled on exit*/
		disable_interrupt(softstate,bus);
	}

	/*Hmmm.. IRQ on udma has been triggered for some reason?? */
	clear_udma_intr(softstate,bus);

	clear_udma_err(softstate,bus);

	DLPRINT(1,"selecting primary drive again");
	ide_select_drive(softstate,0);

	if(debug_level>0){
		printf("STATUS after device probing:");
		printudmastatus(readudmastatus(softstate,bus));
		printstatus(readaltstatus(softstate,bus));
	}


#define PROGUDMA

#ifdef PROGUDMA
	if(primarygood){

		if(init_udma(softstate,0) != 0){
			DLPRINT(1,"hpt: init_udma on primary failed");
			primarygood=0;
		}
		
		hpt_iprint(instance,"Done primary drive setup");
	}
	if(slavegood){

		if(init_udma(softstate,1) != 0){
			DLPRINT(1,"hpt: init_udma on slave failed");
			slavegood=0;
		}
		hpt_iprint(instance,"Done slave drive setup");
	}

	/* XXX This really belongs in a bus-level routine, not a DRIVE
	 * level routine. But I need to test it somehow...
	 */
#ifdef SET_TIMINGS
	printf("DEBUG: hardcoding timings to match drive 0 speed\n");
	hpt_set_timings(softstate,0,softstate->drive_speed[0]);
#endif /* SET_TIMINGS */
	DLPRINT(1,"DEBUG: setting UDMA controller configuration\n");
	hpt_set_controls(softstate);

#endif /* PROGUDMA */

	/* XXXX in future, it SHOULD BE PERMISSIBLE to have nothing
	 * on the main bus, but stuff on the second bus !
	 * Maybe. If there IS a second bus. gah.
	 */
	if((primarygood==0) && (slavegood==0)){
		cmn_err(CE_WARN, "hpt%d: no UDMA drives found ",
			instance);
		printf("DEBUG: here is ide alt status reg");
		printstatus(readaltstatus(softstate,bus));

		cleanup(softstate);
		return DDI_FAILURE;
	}
	
	

	if(primarygood){
		make_minors(dip, instance,0);
	}

	if(slavegood){
		make_minors(dip, instance,1);
	}

	
	DLPRINT(1,"Adding interrupt handler");

	if(ddi_add_intr(dip, 0,
			&softstate->interrupt_cookie,
			0, hpt_handle_intr,
			(caddr_t) instance)
	   != DDI_SUCCESS){
		cmn_err(CE_WARN, "could not add interrupt handler");
		cleanup(softstate);
		return DDI_FAILURE;
	}

#ifdef SOFTINTR_TIMEOUT
	/* for 30-second timeout paranoia */
	if(ddi_add_softintr(dip, DDI_SOFTINT_LOW,
			&softstate->softintr_id,
			NULL,NULL,
			timeout_intr,
			(caddr_t) instance)
	   != DDI_SUCCESS){
		cmn_err(CE_WARN, "could not add soft_interrupt handler");
		cleanup(softstate);
		return DDI_FAILURE;
	}
#endif /*SOFTINTR_TIMEOUT*/


	enable_interrupt(softstate,bus);

	if(got_udma_intr(softstate,bus)){
		DLPRINT(1,"clearing UDMA intr bit");
		clear_udma_intr(softstate,bus);
	}


	hpt_iprint(instance,"hpt_attach returning success");
	
	if(debug_level>1){
		printf("at end of hpt_attach:\n");
		printudmastatus(readudmastatus(softstate,bus));
		printstatus(readaltstatus(softstate,bus));
	}


	/* write something  for dmsg */
	ddi_report_dev(dip);

#ifdef TUNE_REGS
	/* The linux drivers do some undocumented register tuning.
	 * I've tried this, and it doesnt seem to help the lockup problem.
	 */
	reg32=pci_config_get32(softstate->pci_conf_h, 0x40);
	printf("DEBUG: initial check: regspeed is %d\n",(reg32 >> 8) & 7);

	if(udma_max<5){
		reg32=0x10c8a731;
		printf("DEBUG: forcing timing reg to %x\n",reg32);
		pci_config_put32(softstate->pci_conf_h, 0x40,reg32);
		reg32=pci_config_get32(softstate->pci_conf_h, 0x40);
		printf("DEBUG-#2: verify: 0x40 reg is %x\n",reg32);
	}
#endif


	return DDI_SUCCESS;
}

static	int	hpt_open(dev_t *dev, int openflags, int otyp, cred_t *credp){
   	routineprint("open");
	switch(otyp){
	   case OTYP_BLK:
		hpt_print("open on devtype BLK");
		break;
	   case OTYP_CHR:
		hpt_print("open on devtype CHR");
		break;
	   default:
		hpt_print("open on UNKNOWN TYPE??");
		return (ENXIO);
	}

   	/*return (ENXIO);*/
   	return 0;
}

/* modeled after routine in bst.c 
 * This is the shared body of hpt_read() and hpt_write() 
 * We call hpt_strategy(), through physio().
 * physio() will take care of splitting the transfer into multiple
 * calls to hpt_strategy(), if neccessary.
 * It will also lock pages in memory, and handle address mapping to
 * userland somehow. But it's our job to map them to physical addrs for
 * the DMA.
 *
 * note the dependancy on minor-number->instance mapping
 */
static int hpt_rw(dev_t dev, struct uio *uio, int flag){
	struct hpt_softstate *softstate;
	int instance=GETINSTANCE(getminor(dev));
	softstate=ddi_get_soft_state(hpt_globalstate, instance);
	if(softstate == NULL){
		return (EAGAIN);
	}

	DLPRINT2(2,"DEBUG: hpt_rw: iov_len is %ld",
		 uio->uio_iov->iov_len);

	/* has to start on multiple of sector size, for some stupid
	 * law about disk drivers
	 */

	if(uio->uio_offset & (NBPSCTR-1)){
		DLPRINT(1,"hpt_rw: not on sector boundary");
		return EINVAL;
	}
	if(uio->uio_iov->iov_len & (NBPSCTR-1)){
		DLPRINT(1,"hpt_rw: not sectorsize multiple");
		DLPRINT2(1,"  (len of %ld )",uio->uio_iov->iov_len);
		return EINVAL;
	}

	return (physio(hpt_strategy, NULL, dev, flag, minphys, uio));
}

static	int	hpt_read(dev_t dev, struct uio* uiop, cred_t * credp){
   	routineprint("read");

#ifdef ODDHACK
	char *msg="hpt_read test string for buffer copy";
	caddr_t user_addr;

	DLPRINT(1,"DEBUG: hpt_read will be FAKED");
	if(uiop->uio_offset >0){
		cmn_err(CE_WARN,"ERROR: uio_offset != 0");
		return DDI_FAILURE;
	}

	user_addr=uiop->uio_iov->iov_base;
	/* If you want to "cleanly" pass back a string to userland, and
	 * have the uio structures automatically be updated with the
	 * size of data transferred, etc, use uiomove, as follows:
	 */
	/* uiomove(msg,37,UIO_READ, uiop);*/

	/* This is a sneaky way to see if the UDMA transfer is actually
	 * writing over the buffer!!!
	 */
	/*ddi_copyout(msg, user_addr, 37, FREAD); */
#endif /* ODDHACK */

	return (hpt_rw(dev, uiop, B_READ));
}

static	int	hpt_write(dev_t dev, struct uio *uiop, cred_t *credp){
   	routineprint("write");
	return (hpt_rw(dev, uiop, B_WRITE));
}

static	int	hpt_close(dev_t dev, int openflags, int otyp, cred_t *credp){
   	routineprint("close");
	/* we ignore open(), so ignore this too! */
   	return (0);
}

/* This is called by an IOCTL call. After which, we dump
 * various information to console.
 * See hpt_ioctl to find out WHICH one triggers this
 */
void debug_info(struct hpt_softstate *softstate){
	/*int disk=0;*/
	int bus=0;

	mutex_enter(&softstate->bus_mutex[bus]);

/*	printqueue(softstate,disk);*/

	
	printstatus(readaltstatus(softstate,bus));
	printudmastatus(readudmastatus(softstate,bus));
	mutex_exit(&softstate->bus_mutex[bus]);
/*
	DLPRINT2(0," intr_count is %d", intr_count);
	DLPRINT2(0," udmaintr_count is %d", udmaintr_count);
*/

}

/* unstick() is the body of the special IOCTL to 'unstick' the driver.
 * 
 * This ASSUMES that we've somehow left a job on the forward queue, and
 * haven't handled/triggered the interrupt properly somehow.
 *
 * If we've taken it off the active queue, then we're hosed, we're
 * never going to be able to call biodone() on the job.
 *
 */
static void unstick(struct hpt_softstate *softstate,int drive){
	int bus=drive>>1;
	if(softstate->waitqueue[drive].b_forw==NULL){
		printf("unstick: nothing on active queue. Going for reset");

#define UNSTICK_FULLRESET

#ifdef UNSTICK_FULLRESET
		/* REMEMBER: after software reset, need to set drive params!*/
		software_reset(softstate, bus);
		/* reinit BOTH drives on bus*/
		init_udma(softstate,(drive&0xfe));
		init_udma(softstate,(drive&0xfe)|1);
#else
		reset_drive(softstate,drive);
#endif
		
		printf("clearing UDMA IRQ flag, if needed");
		clear_udma_intr(softstate, bus);
		printf("Done with unstick:");
		debug_info(softstate);
		return;
	}
	printf("HPT: unstick of drive %d requested. Calling cleanup_active\n",
		drive);

	hpt_cleanup_active(softstate,drive,EIO);
	
	DLPRINT(1,"unstick done with cleanup");

	if(softstate->waitqueue[drive].av_forw!=NULL){
		DLPRINT(1,"unstick trying to start queue processing again");
		hpt_start_one_job(softstate,drive);
	} else {
		DLPRINT(1,"No more outstanding requests");
	}
	
	DLPRINT(1,"unstick done");
	
}

/* Implements special "read/write raw sector" ioctl.
 * This seems to be to write to the SOLARIS FDISK PARTITION ONLY.
 * NOT the entire disk.
 * Why we need this, when anything can just open up the "c0d0pX" device
 * is beyond me.
 * So, we just pass off to hpt_strategy, after wrapping ioctl
 * memory, with a fake allocated (struct buf)
 */
static int ioctl_rwcmd(struct hpt_softstate *softstate, dev_t dev, int disk,
		       struct dadkio_rwcmd* cmd, int flags)
{
	int retval=0;
	int bus=0;
 	struct dadkio_rwcmd cmd_copy;

#ifdef USE_DMA_FOR_IOCTL
	struct buf*mybuf;
	int sol_fpart;
#endif
	if(disk>1) bus=1;

	ddi_copyin(cmd, &cmd_copy,sizeof(struct dadkio_rwcmd),flags);

	if(cmd->cmd==DADKIO_RWCMD_WRITE){
	printf("DEBUG: ioctl_rwcmd-WRITE: blkaddr = %lu, buflen = %u\n",
	       cmd_copy.blkaddr, cmd_copy.buflen);
	} else {
	printf("DEBUG: ioctl_rwcmd: blkaddr = %lu, buflen = %u\n",
	       cmd_copy.blkaddr, cmd_copy.buflen);
	}

#ifdef USE_DMA_FOR_IOCTL
	sol_fpart=hpt_find_sol_fpart(softstate,disk);
	if(sol_fpart==-1){
		cmn_err(CE_WARN,"ioctl_rwcmd ERROR: no solaris partition exists");
		return EINVAL;
	}

	/* Force "dev" to point to solaris fdisk partition */
	dev&= ~DEV_DRIVEMASK;

	dev|= sol_fpart & DEV_DRIVEMASK;


	if(cmd->cmd==DADKIO_RWCMD_WRITE){
		return 0;
	}

	mybuf=getrbuf(KM_SLEEP);

	switch(cmd->cmd){
		case DADKIO_RWCMD_READ:
			printf("DEBUG: ioctl_rwcmd READ\n");
			mybuf->b_flags=B_READ;
			break;
		case DADKIO_RWCMD_WRITE:
			printf("DEBUG: ioctl_rwcmd WRITE\n");
			mybuf->b_flags=B_WRITE;
			break;
	}

	mybuf->b_lblkno=cmd_copy.blkaddr;
	mybuf->b_un.b_addr=cmd_copy.bufaddr;
	mybuf->b_bcount=cmd_copy.buflen;

	mybuf->b_edev=dev;

	hpt_strategy(mybuf);
	biowait(mybuf);

	retval=geterror(mybuf);
	
	freerbuf(mybuf);

#else /* USE_DMA_ */


	if(cmd->cmd==DADKIO_RWCMD_READ){
		int blockcount=cmd_copy.buflen >> DEV_BSHIFT;
		/* offset so it is inside solaris fdisk partition! */
		int block_start=softstate->sol_startblk[disk] +
				cmd_copy.blkaddr;
		uint8_t *userbuf=cmd_copy.bufaddr;
		uint8_t *tmpbuf;
		int read_stat;

		tmpbuf=kmem_zalloc(blockcount*DEV_BSIZE, KM_SLEEP);

		disable_interrupt(softstate,bus);
		read_stat=hpt_read_pio(softstate,disk,
			    tmpbuf, block_start, blockcount);
			
		if(read_stat != 0)
		{
			cmn_err(CE_WARN,"DADKIO_RWCMD_READ failed");
			retval=EIO;

		} else {
			ddi_copyout(tmpbuf, userbuf,blockcount*DEV_BSIZE,
				    flags);
		}
		enable_interrupt(softstate,bus);
		kmem_free(tmpbuf,blockcount*DEV_BSIZE);
	} else {
		/* WRITE data now... */
		int blockcount=cmd_copy.buflen >> DEV_BSHIFT;
		/* offset so it is inside solaris fdisk partition! */
		int block_start=softstate->sol_startblk[disk] +
				cmd_copy.blkaddr;
		uint8_t *userbuf=cmd_copy.bufaddr;
		uint8_t *tmpbuf;

		tmpbuf=kmem_zalloc(blockcount*DEV_BSIZE, KM_SLEEP);

		ddi_copyin(userbuf, tmpbuf,blockcount*DEV_BSIZE,
			   flags);
		
		disable_interrupt(softstate,bus);
		if(hpt_write_pio(softstate,disk,tmpbuf,block_start,
			     blockcount) != 0)
		{
			cmn_err(CE_WARN,"DADKIO_RWCMD_WRITE failed");
			retval=EIO;
		}
		enable_interrupt(softstate,bus);
		kmem_free(tmpbuf,blockcount*DEV_BSIZE);
	}

#endif

	ddi_copyout(&retval, &(cmd->status).status,sizeof(retval),flags);

	DLPRINT2(2," ioctl_rw returning status of %d",retval);
	/* ioctl expects 0 returned always. Caller checks the status byte
	 * in the struct.
	 */
	return 0;
	
}

/* Most of this ioctl stuff is mandatory because solaris
 * expects certain ioctls to be supported by formattable devices.
 * Otherwise 'format' wont work!
 */
int hpt_ioctl(dev_t dev, int cmd, intptr_t origarg, int flags,
	cred_t *credp, int *rvalp){

	int *arg = (int *)origarg;
	struct hpt_softstate *softstate;
	int minornum=getminor(dev);
	int instance=GETINSTANCE(minornum);
	int drive=(minornum & DEV_DRIVEMASK) / DEV_DRIVEMULT;
	int bus=drive>>1;
	int slice=minornum&DEV_SLICEMASK;

	if(slice>= V_NUMPAR){
		cmn_err(CE_NOTE,"hpt: ioctl: slice number too large");
		return EINVAL;
	}
	
	softstate=ddi_get_soft_state(hpt_globalstate, instance);
	if(softstate == NULL){
		return (EAGAIN);
	}

	DLPRINT2(2,"hpt_ioctl called on slice %d ",slice);

	/**************************************************/
	/* Special hpt-only debug hacks                   */
	/**************************************************/
	switch(cmd){
	    case 1:
		DLPRINT(1,"ioctl_DEBUGINFO");
		debug_info(softstate);
		return 0;
	    case 2:
		DLPRINT(1,"ioctl_UNSTIC");
		unstick(softstate,drive);
		return 0;
	    case 3:
		DLPRINT(1,"ioctl_RESET");
		if(software_reset(softstate,bus)!=0){
			printf("ERROR: software reset of bus %d failed",bus);
		} else {
			DLPRINT2(1,"reset of bus %d suceeded",bus);
		}
		/* Apparently, it is NOT neccessary to reinitialize the
		 * drive after this.. it keeps its parameters
		 */
		return 0;
	    case 4:
		DLPRINT(1,"ioctl_IDENTIFY");
		mutex_enter(&softstate->bus_mutex[bus]);
		
		if(load_identify_info(softstate,drive,1)!=0){
			printf("ERROR: load of info failed?");
		} else {
			print_identity(softstate->dev_info[drive]);
		}
		mutex_exit(&softstate->bus_mutex[bus]);

		return 0;
	} /* End of HACK Cmd switch */

	/* Start of "normal" ioctl handling*/
	switch(cmd){
	    case DIOCTL_RWCMD: /* "direct write" solaris hack */
		DLPRINT(1,"ioctl_RWCMD");
		if((minornum&HPT_FULLDISK) == 0){
			cmn_err(CE_WARN,
				"ioctl(DIOCTL_RWCMD) on non-raw node");
			return EINVAL;
		}
		return ioctl_rwcmd(softstate, dev, drive,
				   (struct dadkio_rwcmd*)origarg, flags);
		break;
		
	    case DKIOCINFO:
	    {
		struct dk_cinfo tmpinfo=controllerinfo;
		hpt_iprint(instance,"ioctl DKIOCINFO");

		/* Note that some values have already been filled out,
		 * in the "controllerinfo" shared struct
		 */
		tmpinfo.dki_cnum=instance;
		tmpinfo.dki_unit=drive;
		tmpinfo.dki_flags=0;
		tmpinfo.dki_partition=dev&0xf;
		/* maxtransfer is also set */

		/* This puts the manufacturer drive name in the field.
		 * Its nice to have... but not the standard as to what
		 * to put in here :-/
		 */
		/*hpt_dname_to_string(softstate->dev_info[drive],
				tmpinfo.dki_dname);
		*/

		ddi_copyout(&tmpinfo, arg, sizeof(tmpinfo),flags);
		return 0;
		
		break;
	    }
	    case DKIOCGGEOM:
		hpt_iprint(instance,"ioctl DKIOCGGEOM");
		{
			struct dk_geom istruct;
			int blkspercyl;
			int pblkcount;
			int solfpnum=hpt_find_sol_fpart(softstate,drive);

			if(solfpnum==-1){
				DLPRINT(1,"no solaris partition for GEOM");
				return EINVAL;
			}


			
			istruct.dkg_nhead = getnumheads(softstate,drive);
			istruct.dkg_nsect = getnumsectpertrack(softstate,drive);
			blkspercyl=istruct.dkg_nhead *istruct.dkg_nsect;

			pblkcount=getfblkcount(softstate,drive,
							solfpnum+1);

			istruct.dkg_pcyl = pblkcount / blkspercyl;

			istruct.dkg_acyl = 2;

			/* I'm not sure we should really "reserve" any
			 * acyl at all. And at any rate, the "backup" slice
			 * spans the WHOLE DISK, and the "alt" stuff as well.
			 * I think the regular solaris drivers return
			 * the "wrong" value anyways,(always losing 2 cyl)
			 * so I'm doing it this way.
			 */
			istruct.dkg_ncyl = istruct.dkg_pcyl;
			istruct.dkg_bcyl = 0;
			istruct.dkg_rpm  = 5400; /* yeah, hardcode */
			
			if(ddi_copyout(&istruct,
			     arg,
			     sizeof(istruct), flags)){
				return EFAULT;
			}
		}

		return 0;


	    case DKIOCG_VIRTGEOM:
	    case DKIOCG_PHYGEOM:
		switch(cmd){
			case DKIOCG_VIRTGEOM:
			hpt_iprint(instance,"ioctl DKIOCG_VIRTGEOM");
			break;
			case DKIOCG_PHYGEOM:
			hpt_iprint(instance,"ioctl DKIOCG_PHYGEOM");
			break;
		}

		{
			struct dk_geom istruct;

			istruct.dkg_nhead = getnumheads(softstate,drive);
			istruct.dkg_nsect = getnumsectpertrack(softstate,drive);
			istruct.dkg_pcyl  = getnumcyl(softstate,drive);
			istruct.dkg_acyl  = 0;
			istruct.dkg_ncyl  = istruct.dkg_pcyl;
			istruct.dkg_bcyl  = 0;
			istruct.dkg_rpm  = 5400;
			
			if(ddi_copyout(&istruct,
			     arg,
			     sizeof(istruct), flags)){
				return EFAULT;
			}
		}

		return 0;/* Good return! */


	    case DKIOCSGEOM:
		hpt_iprint(instance,"ioctl DKIOCSGEOM(faked)");
		return 0; /* good return */
		break;
	    case DKIOCGAPART:
		hpt_iprint(instance,"ioctl DKIOCGAPART(faked)");
		return 0; /* good return */
		break;
	    case DKIOCSAPART:
		hpt_iprint(instance,"ioctl DKIOCSAPART(faked)");
		return 0; /* good return */

	    /* GET solaris partition info */
	    case DKIOCGVTOC:
		hpt_iprint(instance,"ioctl DKIOCGVTOC");

		/* Should return this for ALL DKIOCG calls maybe */
		if(softstate->partitions[drive].v_sanity != VTOC_SANE){
			DLPRINT(1,"  no valid VTOC; returning EIO");
			return EIO;
		}

		if(ddi_copyout(&softstate->partitions[drive],
			       arg,
			       sizeof(struct vtoc), flags)){
			return EFAULT;
		}
		return 0; /* good return! */

	    /* information on this SINGLE partition */
	    case DKIOCPARTINFO:
		hpt_iprint(instance,"ioctl DKIOCPARTINFO");
		if(minornum&HPT_FULLDISK){
			struct part_info tmpstr;
			int fpart=minornum&0x7;

			DLPRINT2(1,"Getting fdisk info for p%d",fpart);
			tmpstr.p_start=getfstartblk(softstate,drive,fpart);
			tmpstr.p_length=getfblkcount(softstate,drive,fpart);
			if(ddi_copyout(&tmpstr, arg,
			       sizeof(struct part_info), flags)){
				return EFAULT;
			}
			return 0;

			
		} else
		{
			struct part_info tmpstr;
			struct partition *pptr;
			pptr=&softstate->partitions[drive].v_part[slice];
			tmpstr.p_start=pptr->p_start;
			tmpstr.p_length=getsblkcount(softstate,drive,slice);
			if(ddi_copyout(&tmpstr, arg,
			       sizeof(struct part_info), flags)){
				return EFAULT;
			}
		}
		return 0; /* good return! */


	    /* SET solaris partition info */
	    case DKIOCSVTOC:
		hpt_iprint(instance,"ioctl DKIOCSVTOC");
		slice=slice&0x7;
		if((minornum&HPT_FULLDISK) ==0){
			cmn_err(CE_WARN,"attempt to set vtoc on non-fdisk devnode");
			return EINVAL;
		}
		if(ddi_copyin(arg, &softstate->partitions[drive],
			       sizeof(struct vtoc), flags)){
			cmn_err(CE_WARN,"Could not copy in DKIOCSVTOC info?");
			return EFAULT;
		}
		if(softstate->fdisklabel[drive][slice].systid != SUNIXOS){
			cmn_err(CE_WARN,"DKIOCSVTOC called on non-solaris fdisk partition.");
			cmn_err(CE_WARN,"Will use solaris one if available.");
		}
		if(hpt_write_vtoc(softstate,dev)!=0){
			DLPRINT(1,"hpt_write_vtoc returned fail");
			return EFAULT;
		}
		if(debug_level>1){
			DLPRINT(1,"New slice layout is...");
			print_sparts(softstate,drive);
		}
		return 0; /* good return! */

	   /* Special damn ioctl just to write to block 0.
	    * Although also gives us a hint to re-evaluate the
	    * fdisk partitions
	    */
	   case DKIOCSMBOOT:
		DLPRINT(1,"ioctl DKIOCSMBOOT");
		{
			struct mboot mtmp;
			if(ddi_copyin(arg, &mtmp,
				      sizeof(struct ipart) *FD_NUMPART,
				      flags)){
				DLPRINT(1,"copyin of DKIOCSMBOOT failed");
				return EFAULT;
			}
			if(copy_fdisk_info(softstate,&mtmp,drive)!=0){
				DLPRINT(1,"copy_fdisk_info returned nonzro");
				return 1;
			} else {
				DLPRINT(1,"copy_fdisk_info returned OK.");
			}
		}

		return 0; /* Good return */
		
	   case DKIOCGMBOOT:
		DLPRINT(1,"ioctl DKIOCGMBOOT called: NOT SUPPORTED YET");
		return ENOTTY;

	   case DKIOCREMOVABLE:
		DLPRINT(1,"ioctl_REMOVABLE");
		{
			int removable=0;
			if(ddi_copyout(&removable, arg,sizeof(int),flags)){
				return EFAULT;
			}
		}
		return 0; /* Good return */

	} /* End of massive ioctl(cmd) switch */

	DLPRINT2(0,"hpt: unknown ioctl %d",cmd);
	/* bst returns ENOTTY. so we will too*/
   	return (ENOTTY);
}

/* Go through all of our disks' wait queues, dequeue, unbind, and
 * bioerror() everything
 *
 * This is called by hpt_detach
 *
 * This could cause major screw-ups if anything is active.
 *
 * So we grab the bus mutexes, and release them when done.
 */
void dequeue_all(struct hpt_softstate *softstate){
	int disk,bus;

	printf("hpt: dequeue_all() called");

	for(bus=0;bus <HPT_NUMBUSES;bus++){
		mutex_enter(&softstate->bus_mutex[bus]);
		/* stop potential DMA transfers in progress!! */
		/* THis should in theory clean up the active queues
		 * as well, since interrupt handler should deal with that.
		 * erm... except that will want the mutex, so
		 * we need to release the bus mutex immediately after
		 * resetting the drive.
		 */
		software_reset(softstate, bus);
		mutex_exit(&softstate->bus_mutex[bus]);
	}

	/* THis is really really bad. we should have SOME kind of protection
	 * here
	 */
	for(disk=0; disk <HPT_NUMDRIVES; disk++){
		hpt_forceclean_waitqueue(softstate,disk);
	}
	for(bus=0;bus <HPT_NUMBUSES;bus++){
	}
	printf("dequeue_all quitting");
	
}

/* This gets called by the system to detach a particular instance
 * of a driver.
 * It is the opposite of hpt_attach().
 *
 */
static	int	hpt_detach(dev_info_t *dip, ddi_detach_cmd_t cmd){
	int instance;
	struct hpt_softstate *softstate;
	
   	routineprint("detach");
	instance= ddi_get_instance(dip);

	switch(cmd){
	   case DDI_DETACH:
		break;
	   case DDI_SUSPEND:
	   default:
		hpt_iprint(instance,"non-DETACH call to hpt_detach: FAIL");
		return (DDI_FAILURE);
	}

	softstate=ddi_get_soft_state(hpt_globalstate, instance);
	if(softstate == NULL){
		hpt_iprint(instance,"detach: no soft state?");
		/* we never really started! */
		return (DDI_SUCCESS);
	}

	DLPRINT(1,"SHOULD call dequeue_all, but cant right now");
	/*
	 * need to have a  "shutting down" flag somewhere
	dequeue_all(softstate);
	*/

	if(softstate->interrupt_cookie != NULL){
		ddi_remove_intr(dip, 0, softstate->interrupt_cookie);
	}
	if(softstate->softintr_id != NULL){
		ddi_remove_softintr(softstate->softintr_id);
	}
	
	cleanup(softstate);
	hpt_iprint(instance, "detach: finished");
	return (DDI_SUCCESS);
		     
}

