#include <fcntl.h>
#include <sys/types.h>
#include <sys/dkio.h>
#include <sys/vtoc.h>
#include <errno.h>

#include <stdio.h>
#include <unistd.h>


void usage(){
	puts("Usage: ");
	puts("  getvtoc /dev/name [block_offset [-i]]");
	puts("     -i means 'ignore bad magic field and print anyway'");
	puts("    NOTE: standard solaris p devs tend to need an offset of 1");
	
}



/* Given a SPECIFIC BLOCK, scan it for vtoc */
/* its up to the caller to do the +1 for solaris irritation */
/* return 0 on fail, 1 on found */
int scanblock(int fd, int offset){
	struct vtoc *label;
	char buffer[1024];
	lseek(fd,offset*512, SEEK_SET);

	if(read(fd,buffer,1024) !=1024){
		perror("err reading initial 1k");
		exit(errno);
	}
	label=(struct vtoc*)buffer;

	if(label->v_sanity == VTOC_SANE){
		if(label->v_part[2].p_start != 0){
			return 0;
		}
		if(label->v_part[8].p_start != 0){
			return 0;
		}
		fprintf(stderr,"FOUND a vtoc at block %d(0x%x)\n",
				offset,offset);

		fprintf(stderr,"volume name %s\n",label->v_volume);

#ifdef NOTNOW
		for(pc=0;pc<10;pc++){
		fprintf(stderr,"%d: startblock %d, size %d, type %d\n",
		pc,
			label->v_part[pc].p_start,
			label->v_part[pc].p_size,
			label->v_part[pc].p_tag);
		}
#endif /* NOTNOW */
		return 1;
	}
}

int main(int argc, char *argv[]){
	int  fd, pc;
	struct vtoc *label;
	int offset, cylsize,cylcount;
	char buffer[1024];

	if(argc<3){
		puts("Need name of  device, and blk count of cylinder!");
		usage();
		exit(1);
	}
	fd=open(argv[1],O_RDONLY);

	if(fd==-1){
		perror("could not open device");
		exit(errno);
	}

	cylsize=atoi(argv[2]);

	for(cylcount=0;cylcount<1500; cylcount++){
		offset=cylcount*cylsize + 1;
		if(scanblock(fd,offset)==1){
			fprintf(stderr,
				"FOUND VTOC at cyl %d (+1 block offset)\n",
			      cylcount);
		}
	}

	return 0;
}
